import argparse
import os  # noqa: F401
import re
import sys  # noqa: F401

from azure.common.client_factory import get_client_from_cli_profile
from azure.mgmt.containerregistry import ContainerRegistryManagementClient
from azureml.core import Datastore, Experiment, Run, Workspace  # noqa: F401
from azureml.core.compute import AmlCompute, ComputeTarget  # noqa: F401
from azureml.core.container_registry import ContainerRegistry
from azureml.core.runconfig import MpiConfiguration, RunConfiguration  # noqa: F401
from azureml.data.azure_storage_datastore import AzureBlobDatastore, AzureFileDatastore  # noqa: F401
from azureml.train.estimator import Estimator

parser = argparse.ArgumentParser()
parser.add_argument(
    "--subscription", type=str, default="ea482afa-3a32-437c-aa10-7de928a9e793"
)  # AI Platform GPU - MLPerf
parser.add_argument(
    "--resource_group", type=str, default="onnx_training", help="Azure resource group containing the AzureML Workspace"
)
parser.add_argument(
    "--workspace", type=str, default="ort_training_dev", help="AzureML Workspace to run the Experiment in"
)
parser.add_argument(
    "--compute_target", type=str, default="onnx-training", help="AzureML Compute target to run the Experiment on"
)
parser.add_argument("--experiment", type=str, default="BERT-ONNX", help="Name of the AzureML Experiment")
parser.add_argument(
    "--tags",
    type=str,
    default=None,
    help="Tags to be added to the submitted run (--tag1=value1 --tag2=value2 --tag3=value3)",
)

parser.add_argument(
    "--datastore", type=str, default="bert_premium", help="AzureML Datastore to be mounted into the Experiment"
)
parser.add_argument(
    "--train_dir", type=str, default="book/train", help="Path in the AzureML Datastore containing the train files"
)
parser.add_argument(
    "--test_dir", type=str, default="book/test", help="Path in the AzureML Datastore containing the test files"
)
parser.add_argument(
    "--train_dir2", type=str, default=None, help="Path in the AzureML Datastore containing the train files for phase 2"
)
parser.add_argument(
    "--test_dir2", type=str, default=None, help="Path in the AzureML Datastore containing the test files for phase 2"
)

parser.add_argument(
    "--container",
    type=str,
    default="onnxtraining.azurecr.io/azureml/bert:latest-openmpi4.0.0-cuda10.1-cudnn7-ubuntu16.04",
    help="Docker container to use to run the Experiment",
)
parser.add_argument(
    "--container_registry_resource_group",
    type=str,
    default="onnx_training",
    help="Azure resource group containing the Azure Container Registry (if not public)",
)

parser.add_argument(
    "--node_count",
    type=int,
    default=1,
    help="Number of nodes to use for the Experiment. If greater than 1, an MPI distributed job will be run.",
)
parser.add_argument(
    "--gpu_count",
    type=int,
    default=1,
    help="Number of GPUs to use per node. If greater than 1, an MPI distributed job will be run.",
)

parser.add_argument(
    "--model_name",
    type=str,
    default="bert_L-24_H-1024_A-16_V_30528_optimized_layer_norm",
    help="Model to be trained (must exist in the AzureML Datastore)",
)
parser.add_argument(
    "--script_params",
    type=str,
    default="",
    help="Training script parameters (--param1=value1 --param2=value2 --param3=value3)",
)
args = parser.parse_args()

# Get the AzureML Workspace to run the Experiment in
ws = Workspace.get(name=args.workspace, subscription_id=args.subscription, resource_group=args.resource_group)

# Get the existing AzureML Compute target
compute_target = ComputeTarget(workspace=ws, name=args.compute_target)

# Get the datastore from current workspace
ds = Datastore.get(workspace=ws, datastore_name=args.datastore)

# Construct common script parameters
script_params = {
    "--model_name": ds.path(args.model_name).as_download(),
    "--train_data_dir": ds.path(args.train_dir).as_mount(),
    "--test_data_dir": ds.path(args.test_dir).as_mount(),
}

# Optional phase2 script parameters
if args.train_dir2:
    script_params["--train_data_dir_phase2"] = ds.path(args.train_dir2).as_mount()
if args.test_dir2:
    script_params["--test_data_dir_phase2"] = ds.path(args.test_dir2).as_mount()

# Allow additional custom script parameters
for params in args.script_params.split(" "):
    key, value = params.split("=")
    script_params[key] = value

# Allow custom tags on the run
tags = {}
if args.tags:
    for tag in args.tags.split(" "):
        key, value = tag.split("=")
        tags[key] = value

# Get container registry information (if private)
container_image = args.container
registry_details = None

acr = re.match("^((\\w+).azurecr.io)/(.*)", args.container)
if acr:
    # Extract the relevant parts from the container image
    #   e.g. onnxtraining.azurecr.io/azureml/bert:latest
    registry_address = acr.group(1)  # onnxtraining.azurecr.io
    registry_name = acr.group(2)  # onnxtraining
    container_image = acr.group(3)  # azureml/bert:latest

    registry_client = get_client_from_cli_profile(ContainerRegistryManagementClient, subscription_id=args.subscription)
    registry_credentials = registry_client.registries.list_credentials(
        args.container_registry_resource_group, registry_name
    )

    registry_details = ContainerRegistry()
    registry_details.address = registry_address
    registry_details.username = registry_credentials.username
    registry_details.password = registry_credentials.passwords[0].value

# MPI configuration if executing a distributed run
mpi = MpiConfiguration()
mpi.process_count_per_node = args.gpu_count

# AzureML Estimator that describes how to run the Experiment
estimator = Estimator(
    source_directory="./",
    script_params=script_params,
    compute_target=compute_target,
    node_count=args.node_count,
    distributed_training=mpi,
    image_registry_details=registry_details,
    use_docker=True,
    custom_docker_image=container_image,
    entry_script="train.py",
    inputs=[ds.path("./").as_mount()],
)

# Start the AzureML Experiment
experiment = Experiment(workspace=ws, name=args.experiment)
run = experiment.submit(estimator, tags)
print(f"Experiment running at: {run.get_portal_url()}")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

from __future__ import annotations

import ctypes

import torch
from onnx import ModelProto, NodeProto, TensorProto, helper

from onnxruntime.training.ortmodule._pythonop_helper import make_pythonop_node
from onnxruntime.training.utils import onnx_dtype_to_pytorch_dtype, pytorch_type_to_onnx_dtype

MEM_EFFICIENT_PARAM_TRIGGER_INPUT_NAME = "mem_efficient_pull_weight_trigger"
MEM_EFFICIENT_PARAM_TRIGGER_OUTPUT_DTYPE = TensorProto.FLOAT
MEM_EFFICIENT_PARAM_TRIGGER_OUTPUT_SHAPE = [1]


def get_params_connected_to_pull_param_trigger(
    named_params: dict[str, torch.nn.parameter.Parameter], exported_model: ModelProto
):
    # Be noted, some parameters might not in graph input because they are not used in forward, so we filtered them also.
    onnx_initializer_names = {p.name for p in exported_model.graph.input}
    return {k: v for k, v in named_params if v.requires_grad and k in onnx_initializer_names}


def post_processing_enable_mem_efficient_training(
    exported_model: ModelProto,
    named_params: dict[str, torch.nn.parameter.Parameter],
    device: torch.device,
) -> tuple[bool, ModelProto, torch.Tensor]:
    """This function is used to enable memory efficient gradient management.

    Args:
        exported_model: The exported model.
        named_params: The full parameter map.
        device: The device to run the model.

    Returns:
        A tuple of bool, ModelProto and param_trigger_grad tensor. The bool indicates whether the model is modified.

    """
    trainable_named_params = get_params_connected_to_pull_param_trigger(named_params, exported_model)
    if len(trainable_named_params) == 0:
        return False, exported_model, None

    # Create weight retrieving function using trainable_named_params.
    param_trigger_grad = torch.zeros(
        MEM_EFFICIENT_PARAM_TRIGGER_OUTPUT_SHAPE,
        dtype=onnx_dtype_to_pytorch_dtype(MEM_EFFICIENT_PARAM_TRIGGER_OUTPUT_DTYPE),
        device=device,
    )
    param_retrieve_func_class = _create_param_retrieval_function(trainable_named_params, param_trigger_grad)

    graph_inputs_to_remove = []
    input_offset = 0
    for graph_input in exported_model.graph.input:
        if graph_input.name not in trainable_named_params:
            continue

        graph_inputs_to_remove.append(graph_input)

        # Create the param retrieval function for this parameter.
        node_inputs = [
            helper.make_tensor_value_info(
                MEM_EFFICIENT_PARAM_TRIGGER_INPUT_NAME,
                MEM_EFFICIENT_PARAM_TRIGGER_OUTPUT_DTYPE,
                MEM_EFFICIENT_PARAM_TRIGGER_OUTPUT_SHAPE,
            ),
            graph_input.name,  # Second param is a string, which represents the param_name
        ]

        node_outputs = [
            helper.make_tensor_value_info(
                graph_input.name,  # output use the same name as weight
                int(pytorch_type_to_onnx_dtype(trainable_named_params[graph_input.name].dtype)),
                list(trainable_named_params[graph_input.name].shape),
            ),
        ]

        new_node = make_pythonop_node(
            f"weight_retrieval_{graph_input.name}",
            node_inputs,
            node_outputs,
            param_retrieve_func_class,
            training_mode=1,
            safe_run_mode=0,
        )
        exported_model.graph.node.insert(input_offset, new_node)
        input_offset += 1

    # Delete exported_model.graph.input
    names_to_remove = [input.name for input in graph_inputs_to_remove]
    value_infos_to_remove = [
        value_info for value_info in exported_model.graph.value_info if value_info.name in names_to_remove
    ]
    for value_info in value_infos_to_remove:
        exported_model.graph.value_info.remove(value_info)

    for input_to_remove in graph_inputs_to_remove:
        exported_model.graph.input.remove(input_to_remove)

    # Re-order graph input to make sure the weight pull trigger is the first user input.
    offset = 0  # Find the first trainable param, and insert the new input before it, as part of user inputs.
    for input in exported_model.graph.input:
        if input.name in named_params:
            break
        offset += 1
    exported_model.graph.input.insert(
        offset,
        helper.make_tensor_value_info(
            MEM_EFFICIENT_PARAM_TRIGGER_INPUT_NAME,
            MEM_EFFICIENT_PARAM_TRIGGER_OUTPUT_DTYPE,  # Use the same data type with output for the input
            MEM_EFFICIENT_PARAM_TRIGGER_OUTPUT_SHAPE,
        ),
    )

    return True, exported_model, param_trigger_grad


_PARAM_FUNCTION_INDEX = [0]


def _create_param_retrieval_function(
    trainable_named_params: dict[str, torch.nn.parameter.Parameter], param_trigger: torch.Tensor
):
    """This function is used to create a weight retrieving function using trainable_named_params.

    Args:
        trainable_named_params: The trainable named parameters.
        param_trigger: The trigger tensor for pulling the weights. param_trigger is pre-allocated just once
            before model execution, later it will be reused by each iteration. This could save the unnecessary
            overhead allocating for each iteration run.

    """

    @staticmethod
    def forward(ctx, param_trigger, param_name):
        ctx.param_name = param_name
        return trainable_named_params[param_name]

    @staticmethod
    def backward(ctx, *grad_outputs):
        trainable_named_params[ctx.param_name].backward(grad_outputs[0])
        return param_trigger, None

    @staticmethod
    def infer_shape(
        node: NodeProto,
        tensor_input_shapes: list[list[int | str] | None],
        tensor_input_dtypes: list[torch.onnx.TensorProtoDataType],
    ) -> tuple[list[list[int | str] | None], list[torch.onnx.TensorProtoDataType]]:
        input_pointer_scalars_attr_name = "input_pointer_scalars"
        found = [attr for attr in node.attribute if attr.name == input_pointer_scalars_attr_name]

        assert len(found) == 1
        input_pointer_scalars = found[0].ints

        # Restore the nn.Module from the pointer.
        param_name = ctypes.cast(input_pointer_scalars[0], ctypes.py_object).value

        tensor_output_shapes = [
            list(trainable_named_params[param_name].shape),
        ]
        tensor_output_dtypes = [
            int(pytorch_type_to_onnx_dtype(trainable_named_params[param_name].dtype)),
        ]

        return tensor_output_shapes, tensor_output_dtypes

    _PARAM_FUNCTION_INDEX[0] += 1

    return type(
        f"ParamRetrievalFunction_{_PARAM_FUNCTION_INDEX[0]}",
        (torch.autograd.Function,),
        {
            "forward": forward,
            "backward": backward,
            "infer_shape": infer_shape,
        },
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import re
from typing import Any

import sympy


def extract_shape_from_symbol(symbol: str) -> int:
    match = re.match(r"i(\d+)_dim(\d+)_(\d+)", symbol)
    assert match
    return int(match.group(3))


def sympy_dot(seq1: list[sympy.Expr], seq2: list[sympy.Expr]) -> sympy.Expr:
    assert len(seq1) == len(seq2)
    return sympy.expand(sum(a * b for a, b in zip(seq1, seq2, strict=False)))


def parse_shape(shape: list[Any]) -> list[sympy.Expr]:
    symbol_shapes = []
    for dim in shape:
        symbol_dim = dim
        if isinstance(dim, str):
            symbol_dim = sympy.Symbol(re.sub(r"[^a-zA-Z0-9_]+", "_", dim))
        elif isinstance(dim, int):
            symbol_dim = sympy.Integer(dim)
        symbol_shapes.append(symbol_dim)
    return symbol_shapes



import argparse

import pandas as pd


def parse_arguments():
    # create parser
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--prev", required=True, help="previous csv")
    parser.add_argument("-c", "--current", required=True, help="current csv")
    parser.add_argument("-o", "--output_csv", required=True, help="output different csv")
    args = parser.parse_args()
    return args


def main():
    args = parse_arguments()
    a = pd.read_csv(args.prev)
    b = pd.read_csv(args.current)
    common = b.merge(a, on=["model", "ep", "error type", "error message"])
    diff = (
        b.append(common, ignore_index=True)
        .drop_duplicates(["model", "ep", "error type", "error message"], keep=False)
        .loc[: b.index.max()]
    )
    diff.to_csv(args.output_csv)


if __name__ == "__main__":
    main()



import argparse
import os

import numpy
import psutil
from onnx import TensorProto

"""
This profiler tool could run a transformer model and print out the kernel time spent on each Node of the model.
Example of profiling of longformer model:
    python profiler.py --model longformer-base-4096_fp32.onnx --batch_size 1 --sequence_length 4096 --global_length 8 --samples 1000 --thread_num 8 --dummy_inputs longformer --use_gpu
Example of importing profile result file from onnxruntime_perf_test:
    python profiler.py --input profile_2021-10-25_12-02-41.json
"""


def parse_arguments(argv=None):
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-i",
        "--input",
        required=False,
        type=str,
        help="Set the input file for reading the profile results",
    )

    parser.add_argument(
        "-m",
        "--model",
        required=False,
        type=str,
        help="onnx model path to run profiling. Required when --input is not specified.",
    )

    parser.add_argument(
        "-b",
        "--batch_size",
        required=False,
        type=int,
        default=1,
        help="batch size of input",
    )

    parser.add_argument(
        "-s",
        "--sequence_length",
        required=False,
        type=int,
        default=32,
        help="sequence length of input",
    )

    parser.add_argument(
        "--past_sequence_length",
        required=False,
        type=int,
        default=1,
        help="past sequence length for gpt2",
    )

    parser.add_argument(
        "--global_length",
        required=False,
        type=int,
        default=1,
        help="number of global tokens for longformer",
    )

    parser.add_argument(
        "--samples",
        required=False,
        type=int,
        default=1000,
        help="number of samples to test. Set it large enough to reduce the variance of performance result.",
    )

    parser.add_argument(
        "--threshold",
        required=False,
        type=float,
        default=0.01,
        help="Threshold of run time ratio among all nodes. Nodes with larger ratio will show in top expensive nodes.",
    )

    parser.add_argument(
        "--thread_num",
        required=False,
        type=int,
        default=-1,
        help="number of threads to use",
    )

    parser.add_argument(
        "--input_ids_name",
        required=False,
        type=str,
        default=None,
        help="input name for input IDs, for bert",
    )
    parser.add_argument(
        "--segment_ids_name",
        required=False,
        type=str,
        default=None,
        help="input name for segment IDs, for bert",
    )
    parser.add_argument(
        "--input_mask_name",
        required=False,
        type=str,
        default=None,
        help="input name for attention mask, for bert",
    )

    parser.add_argument(
        "--dummy_inputs",
        required=False,
        default="default",
        choices=["bert", "gpt2", "longformer", "default"],
        help="Type of model inputs. The default will create dummy inputs with ones.",
    )

    parser.add_argument("-g", "--use_gpu", required=False, action="store_true", help="use GPU")
    parser.set_defaults(use_gpu=False)

    parser.add_argument(
        "--provider",
        required=False,
        type=str,
        default="cuda",
        help="Execution provider to use",
    )

    parser.add_argument(
        "--basic_optimization",
        required=False,
        action="store_true",
        help="Enable only basic graph optimizations. By default, all optimizations are enabled in OnnxRuntime",
    )
    parser.set_defaults(basic_optimization=False)

    parser.add_argument(
        "--kernel_time_only",
        required=False,
        action="store_true",
        help="Only include the kernel time and no fence time",
    )
    parser.set_defaults(kernel_time_only=False)

    parser.add_argument("-v", "--verbose", required=False, action="store_true")
    parser.set_defaults(verbose=False)

    return parser.parse_args(argv)


def run_profile(onnx_model_path, use_gpu, provider, basic_optimization, thread_num, all_inputs):
    from benchmark_helper import create_onnxruntime_session

    session = create_onnxruntime_session(
        onnx_model_path,
        use_gpu,
        provider,
        enable_all_optimization=not basic_optimization,
        num_threads=thread_num,
        enable_profiling=True,
    )

    for inputs in all_inputs:
        _ = session.run(None, inputs)

    profile_file = session.end_profiling()
    return profile_file


def get_dim_from_type_proto(dim):
    return getattr(dim, dim.WhichOneof("value")) if type(dim.WhichOneof("value")) == str else None  # noqa: E721


def get_shape_from_type_proto(type_proto):
    return [get_dim_from_type_proto(d) for d in type_proto.tensor_type.shape.dim]


def create_dummy_inputs(onnx_model, batch_size, sequence_length, samples):
    """Create dummy inputs for ONNX model.

    Args:
        onnx_model (OnnxModel): ONNX model
        batch_size (int): batch size
        sequence_length (int): sequence length
        samples (int): number of samples

    Returns:
        List[Dict]: list of inputs
    """
    dummy_inputs = {}
    for graph_input in onnx_model.get_graph_inputs_excluding_initializers():
        shape = get_shape_from_type_proto(graph_input.type)
        symbol_dims = []
        for i, dim in enumerate(shape):
            if isinstance(dim, str):
                symbol_dims.append(i)

        # allowed symbolic dimensions: batch_size and sequence_length
        if len(symbol_dims) > 2:
            return None
        if len(symbol_dims) > 0:
            shape[symbol_dims[0]] = batch_size
        if len(symbol_dims) > 1:
            shape[symbol_dims[1]] = sequence_length

        elem_type = graph_input.type.tensor_type.elem_type
        assert elem_type in [TensorProto.FLOAT, TensorProto.INT32, TensorProto.INT64]
        data_type = (
            numpy.float32
            if elem_type == TensorProto.FLOAT
            else (numpy.int64 if elem_type == TensorProto.INT64 else numpy.int32)
        )
        data = numpy.ones(shape, dtype=data_type)
        dummy_inputs[graph_input.name] = data

    all_inputs = [dummy_inputs for _ in range(samples)]
    return all_inputs


def create_bert_inputs(
    onnx_model,
    batch_size,
    sequence_length,
    samples,
    input_ids_name=None,
    segment_ids_name=None,
    input_mask_name=None,
):
    """Create dummy inputs for BERT model.

    Args:
        onnx_model (OnnxModel): ONNX model
        batch_size (int): batch size
        sequence_length (int): sequence length
        samples (int): number of samples
        input_ids_name (str, optional): Name of graph input for input IDs. Defaults to None.
        segment_ids_name (str, optional): Name of graph input for segment IDs. Defaults to None.
        input_mask_name (str, optional): Name of graph input for attention mask. Defaults to None.

    Returns:
        List[Dict]: list of inputs
    """
    from bert_test_data import find_bert_inputs, generate_test_data

    input_ids, segment_ids, input_mask = find_bert_inputs(onnx_model, input_ids_name, segment_ids_name, input_mask_name)
    all_inputs = generate_test_data(
        batch_size,
        sequence_length,
        test_cases=samples,
        seed=123,
        verbose=False,
        input_ids=input_ids,
        segment_ids=segment_ids,
        input_mask=input_mask,
        random_mask_length=False,
    )

    return all_inputs


def create_gpt2_inputs(onnx_model, batch_size, sequence_length, past_sequence_length, samples):
    """Create dummy inputs for GPT-2 model.

    Args:
        onnx_model (OnnxModel): ONNX model
        batch_size (int): batch size
        sequence_length (int): sequence length
        past_sequence_length (int): past sequence length
        samples (int): number of samples

    Raises:
        RuntimeError: symbolic is not supported. Use the tool convert_to_onnx.py to export ONNX model instead.

    Returns:
        List[Dict]: list of inputs
    """
    # The symbolic names shall be same as those used in Gpt2Helper.export_onnx(...) function.
    symbols = {
        "batch_size": batch_size,
        "seq_len": sequence_length,
        "past_seq_len": past_sequence_length,
        "total_seq_len": sequence_length + past_sequence_length,
    }

    dummy_inputs = {}
    for graph_input in onnx_model.get_graph_inputs_excluding_initializers():
        shape = get_shape_from_type_proto(graph_input.type)
        for i, dim in enumerate(shape):
            if isinstance(dim, str):
                if dim not in symbols:
                    raise RuntimeError(f"symbol is not supported: {dim}")
                else:
                    shape[i] = symbols[dim]

        elem_type = graph_input.type.tensor_type.elem_type
        assert elem_type in [TensorProto.FLOAT, TensorProto.INT32, TensorProto.INT64]
        data_type = (
            numpy.float32
            if elem_type == TensorProto.FLOAT
            else (numpy.int64 if elem_type == TensorProto.INT64 else numpy.int32)
        )
        data = numpy.ones(shape, dtype=data_type)
        dummy_inputs[graph_input.name] = data

    all_inputs = [dummy_inputs for _ in range(samples)]
    return all_inputs


def create_longformer_inputs(onnx_model, batch_size, sequence_length, global_length, samples):
    """Create dummy inputs for Longformer model.

    Args:
        onnx_model (OnnxModel): ONNX model
        batch_size (int): batch size
        sequence_length (int): sequence length
        global_length (int): number of global tokens
        samples (int): number of samples

    Raises:
        RuntimeError: symbolic is not supported. Use the tool convert_longformer_to_onnx.py to export ONNX model instead.

    Returns:
        List[Dict]: list of inputs
    """
    symbols = {"batch_size": batch_size, "sequence_length": sequence_length}

    dummy_inputs = {}
    for graph_input in onnx_model.get_graph_inputs_excluding_initializers():
        shape = get_shape_from_type_proto(graph_input.type)
        for i, dim in enumerate(shape):
            if isinstance(dim, str):
                if dim not in symbols:
                    raise RuntimeError(f"symbol is not supported: {dim}")
                else:
                    shape[i] = symbols[dim]

        elem_type = graph_input.type.tensor_type.elem_type
        assert elem_type in [TensorProto.FLOAT, TensorProto.INT32, TensorProto.INT64]
        data_type = (
            numpy.float32
            if elem_type == TensorProto.FLOAT
            else (numpy.int64 if elem_type == TensorProto.INT64 else numpy.int32)
        )

        if "global" in graph_input.name:
            data = numpy.zeros(shape, dtype=data_type)
            data[:, :global_length] = 1
        else:
            data = numpy.ones(shape, dtype=data_type)
        dummy_inputs[graph_input.name] = data

    all_inputs = [dummy_inputs for _ in range(samples)]
    return all_inputs


def run(args):
    num_threads = args.thread_num if args.thread_num > 0 else psutil.cpu_count(logical=False)

    # Set OMP environment variable before importing onnxruntime. Needed for cpu only, and no impact for onnxruntime-gpu package.
    if "OMP_NUM_THREADS" not in os.environ:
        os.environ["OMP_NUM_THREADS"] = str(num_threads)

    from onnx import load
    from onnx_model import OnnxModel

    onnx_model = OnnxModel(load(args.model))

    all_inputs = None
    if args.dummy_inputs == "bert":
        all_inputs = create_bert_inputs(
            onnx_model,
            args.batch_size,
            args.sequence_length,
            args.samples,
            args.input_ids_name,
            args.segment_ids_name,
            args.input_mask_name,
        )
    elif args.dummy_inputs == "gpt2":
        all_inputs = create_gpt2_inputs(
            onnx_model,
            args.batch_size,
            args.sequence_length,
            args.past_sequence_length,
            args.samples,
        )
    elif args.dummy_inputs == "longformer":
        all_inputs = create_longformer_inputs(
            onnx_model,
            args.batch_size,
            args.sequence_length,
            args.global_length,
            args.samples,
        )
    else:  # default
        all_inputs = create_dummy_inputs(onnx_model, args.batch_size, args.sequence_length, args.samples)

    profile_file = run_profile(
        args.model,
        args.use_gpu,
        args.provider,
        args.basic_optimization,
        args.thread_num,
        all_inputs,
    )

    return profile_file


if __name__ == "__main__":
    arguments = parse_arguments()
    print("Arguments", arguments)

    from benchmark_helper import setup_logger

    setup_logger(arguments.verbose)

    if not arguments.input:
        assert arguments.model, "requires either --model to run profiling or --input to read profiling results"
        profile_file = run(arguments)
    else:
        profile_file = arguments.input
    from profile_result_processor import process_results

    results = process_results(profile_file, arguments)

    for line in results:
        print(line)



# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import sys
import unittest

import numpy as np
import onnx
from onnx import TensorProto, helper
from parity_utilities import find_transformers_source

if find_transformers_source():
    from fusion_options import FusionOptions
    from onnx_model import OnnxModel
    from optimizer import optimize_model
else:
    from onnxruntime.transformers.fusion_options import FusionOptions
    from onnxruntime.transformers.onnx_model import OnnxModel
    from onnxruntime.transformers.optimizer import optimize_model


def float_tensor(name: str, shape: list[int], random=False):
    low = 0.0
    high = 1.0
    total_elements = 1
    for x in shape:
        total_elements *= x
    weights = [np.random.uniform(low, high) for _ in range(total_elements)] if random else [1.0] * total_elements
    return helper.make_tensor(name, TensorProto.FLOAT, shape, weights)


class TestRotaryEmbeddingFusion(unittest.TestCase):
    def setUp(self):
        self.batch_size = 2
        self.sequence_length = 8
        self.num_heads = 4
        self.head_size = 6
        self.hidden_size = self.num_heads * self.head_size

        self.past_sequence_length = 2
        self.max_sequence_length = 12

    def verify_fusion(self, expected_model_path, original_model_path):
        expected_model = OnnxModel(onnx.load(expected_model_path))
        expected_model.topological_sort(is_deterministic=True)

        options = FusionOptions("gpt2")
        optimized_model = optimize_model(original_model_path, optimization_options=options, opt_level=0)
        optimized_model.topological_sort(is_deterministic=True)

        self.assertTrue(str(expected_model.model.graph), str(optimized_model.model.graph))

    def create_initializers(self):
        initializers = [
            float_tensor("cos_cache", [self.max_sequence_length, self.head_size]),
            float_tensor("sin_cache", [self.max_sequence_length, self.head_size]),
            helper.make_tensor(
                "pos_ids_new_shape",
                TensorProto.FLOAT,
                [2],
                np.array([self.batch_size, self.sequence_length], dtype=np.int64),
            ),
            helper.make_tensor("zero", TensorProto.FLOAT, [1], np.array([0], dtype=np.int64)),
            helper.make_tensor("one", TensorProto.FLOAT, [1], np.array([1], dtype=np.int64)),
            helper.make_tensor("two", TensorProto.FLOAT, [1], np.array([2], dtype=np.int64)),
            helper.make_tensor("three", TensorProto.FLOAT, [1], np.array([3], dtype=np.int64)),
            helper.make_tensor("int_max", TensorProto.FLOAT, [1], np.array([sys.maxsize], dtype=np.int64)),
        ]
        return initializers

    def create_inputs_and_outputs(self, model_type: str = ""):
        inputs = [
            helper.make_tensor_value_info(
                "input_0",
                TensorProto.FLOAT,
                [self.batch_size, self.sequence_length, self.num_heads, self.head_size],
            ),
            helper.make_tensor_value_info("position_ids", TensorProto.INT64, [self.batch_size, self.sequence_length]),
        ]
        if model_type in {"past", "merged"}:
            # Input will be removed in fused model since it's not used in RotaryEmbedding.
            # We create this input so that we can check the `past_seq_len` path during
            # RotaryEmbedding fusion.
            inputs.append(
                helper.make_tensor_value_info(
                    "past_key",
                    TensorProto.FLOAT,
                    [self.batch_size, self.num_heads, self.past_sequence_length, self.head_size],
                )
            )
        # Dummy input to test nodes for `curr_seq_len` path
        if model_type != "":
            inputs.append(
                helper.make_tensor_value_info(
                    "curr_key",
                    TensorProto.FLOAT,
                    [self.batch_size, self.sequence_length, self.num_heads, self.head_size],
                )
            )
        outputs = [
            helper.make_tensor_value_info(
                "output_0",
                TensorProto.FLOAT,
                [self.batch_size, self.num_heads, self.sequence_length, self.head_size],
            )
        ]
        if model_type in {"merged"}:
            # Dummy output to test that nodes for `past_seq_len` path are not removed for merged model
            outputs.append(helper.make_tensor_value_info("past_seq_len_plus_zero", TensorProto.FLOAT, [1]))
        return inputs, outputs

    def create_fused_model(self, interleaved: bool, initializers: list[TensorProto]):
        inputs, outputs = self.create_inputs_and_outputs()

        rope_node = helper.make_node(
            "RotaryEmbedding",
            inputs=[inputs[0].name, inputs[1].name, initializers[0].name, initializers[1].name],
            outputs=[outputs[0].name],
            name="RotaryEmbedding_0",
            interleaved=int(interleaved),
        )

        graph = helper.make_graph(
            nodes=[rope_node],
            name="RotaryEmbedding_Graph",
            inputs=inputs,
            outputs=outputs,
            initializer=initializers,
        )
        opset_import = helper.make_opsetid(domain="com.microsoft", version=1)
        model = helper.make_model(graph, opset_imports=[opset_import])
        return model

    def create_cache_path(self, model_type: str, use_redundant_squeeze_ops: bool):
        # Create position ids path
        reshape_node = helper.make_node(
            "Reshape",
            inputs=["position_ids", "pos_ids_new_shape"],
            outputs=["pos_ids_reshaped"],
            name="Reshape_0",
        )
        pos_ids_nodes = [reshape_node]

        # Create cos path
        cos_init_unsqueeze_node = helper.make_node(
            "Unsqueeze",
            inputs=["new_seq_len", "zero"],
            outputs=["cos_unsqueeze"],
            name="Unsqueeze_2",
        )
        cos_slice_node = helper.make_node(
            "Slice",
            inputs=["cos_cache", "zero", "cos_unsqueeze", "two", "one"],
            outputs=["cos_sliced"],
            name="Slice_2",
        )
        cos_nodes = [cos_init_unsqueeze_node, cos_slice_node]

        if use_redundant_squeeze_ops:
            # These two nodes are eliminated by this transformers PR: https://github.com/huggingface/transformers/pull/26162
            cos_squeeze_1_node = helper.make_node(
                "Squeeze",
                inputs=["cos_sliced", "zero"],
                outputs=["cos_squeeze_1"],
                name="Squeeze_0",
            )
            cos_squeeze_2_node = helper.make_node(
                "Squeeze",
                inputs=["cos_squeeze_1", "zero"],
                outputs=["cos_squeeze_2"],
                name="Squeeze_1",
            )
            cos_nodes.extend([cos_squeeze_1_node, cos_squeeze_2_node])

        cos_gather_node = helper.make_node(
            "Gather",
            inputs=["cos_squeeze_2" if use_redundant_squeeze_ops else "cos_sliced", "pos_ids_reshaped"],
            outputs=["cos_indexed"],
            name="Gather_1",
        )
        cos_end_unsqueeze_node = helper.make_node(
            "Unsqueeze",
            inputs=["cos_indexed", "one"],
            outputs=["cos"],
            name="Unsqueeze_3",
        )
        cos_nodes.extend([cos_gather_node, cos_end_unsqueeze_node])

        # Create sin path
        sin_init_unsqueeze_node = helper.make_node(
            "Unsqueeze",
            inputs=["new_seq_len", "zero"],
            outputs=["sin_unsqueeze"],
            name="Unsqueeze_4",
        )
        sin_slice_node = helper.make_node(
            "Slice",
            inputs=["sin_cache", "zero", "sin_unsqueeze", "two", "one"],
            outputs=["sin_sliced"],
            name="Slice_3",
        )
        sin_nodes = [sin_init_unsqueeze_node, sin_slice_node]

        if use_redundant_squeeze_ops:
            sin_squeeze_1_node = helper.make_node(
                "Squeeze",
                inputs=["sin_sliced", "zero"],
                outputs=["sin_squeeze_1"],
                name="Squeeze_2",
            )
            sin_squeeze_2_node = helper.make_node(
                "Squeeze",
                inputs=["sin_squeeze_1", "zero"],
                outputs=["sin_squeeze_2"],
                name="Squeeze_3",
            )
            sin_nodes.extend([sin_squeeze_1_node, sin_squeeze_2_node])

        sin_gather_node = helper.make_node(
            "Gather",
            inputs=["sin_squeeze_2" if use_redundant_squeeze_ops else "sin_sliced", "pos_ids_reshaped"],
            outputs=["sin_indexed"],
            name="Gather_2",
        )
        sin_end_unsqueeze_node = helper.make_node(
            "Unsqueeze",
            inputs=["sin_indexed", "one"],
            outputs=["sin"],
            name="Unsqueeze_5",
        )
        sin_nodes.extend([sin_gather_node, sin_end_unsqueeze_node])

        # Create beginning nodes before cos and sin paths

        # Create curr seq len path
        curr_transpose_node = helper.make_node(
            "Transpose",
            inputs=["curr_key"],
            outputs=["curr_key_transposed"],
            name="Transpose_curr",
            perm=[0, 2, 1, 3],
        )
        curr_shape_node = helper.make_node(
            "Shape",
            inputs=["curr_key_transposed"],
            outputs=["curr_shape"],
            name="Shape_curr",
        )
        curr_gather_node = helper.make_node(
            "Gather",
            inputs=["curr_shape", "two"],
            outputs=["curr_seq_len" if model_type in {"past", "merged"} else "new_seq_len"],
            name="Gather_curr",
        )
        beginning_nodes = [curr_transpose_node, curr_shape_node, curr_gather_node]

        if model_type in {"past", "merged"}:
            # Create past seq len path
            past_shape_node = helper.make_node(
                "Shape",
                inputs=["past_key"],
                outputs=["past_shape"],
                name="Shape_past",
            )
            past_gather_node = helper.make_node(
                "Gather",
                inputs=["past_shape", "two"],
                outputs=["past_seq_len"],
                name="Gather_past",
            )
            add_node = helper.make_node(
                "Add",
                inputs=["curr_seq_len", "past_seq_len"],
                outputs=["new_seq_len"],
                name="Add_1",
            )
            beginning_nodes.extend([past_shape_node, past_gather_node, add_node])

        if model_type == "merged":
            dummy_node = helper.make_node(
                "Add",
                inputs=["past_seq_len", "zero"],
                outputs=["past_seq_len_plus_zero"],
                name="Add_dummy_node",
            )
            beginning_nodes.append(dummy_node)

        return pos_ids_nodes + cos_nodes + sin_nodes + beginning_nodes

    def create_apply_rope_path(self):
        start_node = helper.make_node(
            "Transpose",
            inputs=["input_0"],
            outputs=["x"],
            name="Transpose_0",
            perm=[0, 2, 1, 3],
        )

        # Calculate x_half_shape
        shape_node = helper.make_node(
            "Shape",
            inputs=["x"],
            outputs=["x_shape"],
            name="Shape_0",
        )
        gather_node = helper.make_node(
            "Gather",
            inputs=["x_shape", "three"],
            outputs=["x_last_idx_shape"],
            name="Gather_0",
            axis=0,
        )
        div_node = helper.make_node(
            "Div",
            inputs=["x_last_idx_shape", "two"],
            outputs=["x_half_shape"],
            name="Div_0",
        )
        unsqueeze_0_node = helper.make_node(
            "Unsqueeze",
            inputs=["x_half_shape", "zero"],
            outputs=["x_half_shape_0"],
            name="Unsqueeze_0",
        )
        unsqueeze_1_node = helper.make_node(
            "Unsqueeze",
            inputs=["x_half_shape", "zero"],
            outputs=["x_half_shape_1"],
            name="Unsqueeze_1",
        )
        x_half_shape_nodes = [shape_node, gather_node, div_node, unsqueeze_0_node, unsqueeze_1_node]

        # Calculate rotate_half
        x1_node = helper.make_node(
            "Slice",
            inputs=["x", "zero", "x_half_shape_0", "three", "one"],
            outputs=["x1"],
            name="Slice_0",
        )
        x2_node = helper.make_node(
            "Slice",
            inputs=["x", "x_half_shape_1", "int_max", "three", "one"],
            outputs=["x2"],
            name="Slice_1",
        )
        neg_node = helper.make_node(
            "Neg",
            inputs=["x2"],
            outputs=["x2_neg"],
            name="Neg_0",
        )
        x_rotate_half_node = helper.make_node(
            "Concat",
            inputs=["x2_neg", "x1"],
            outputs=["x_rotate_half"],
            name="Concat_0",
            axis=-1,
        )
        rotate_half_nodes = [x1_node, x2_node, neg_node, x_rotate_half_node]

        # Calculate x_embed
        x_cos_node = helper.make_node(
            "Mul",
            inputs=["x", "cos"],
            outputs=["x_cos"],
            name="Mul_0",
        )
        x_sin_node = helper.make_node(
            "Mul",
            inputs=["x_rotate_half", "sin"],
            outputs=["x_rotate_half_sin"],
            name="Mul_1",
        )
        end_node = helper.make_node(
            "Add",
            inputs=["x_cos", "x_rotate_half_sin"],
            outputs=["output_0"],
            name="Add_0",
        )
        x_embed_nodes = [start_node, x_cos_node, x_sin_node, end_node]

        return x_half_shape_nodes + rotate_half_nodes + x_embed_nodes

    def create_test_model(self, model_type: str, use_redundant_squeeze_ops: bool, initializers: list[TensorProto]):
        apply_rope_nodes = self.create_apply_rope_path()
        cache_nodes = self.create_cache_path(model_type, use_redundant_squeeze_ops)
        inputs, outputs = self.create_inputs_and_outputs(model_type)

        graph = helper.make_graph(
            nodes=apply_rope_nodes + cache_nodes,
            name="RotaryEmbedding_Graph",
            inputs=inputs,
            outputs=outputs,
            initializer=initializers,
        )
        opset_import = helper.make_opsetid(domain="ai.onnx", version=13)
        model = helper.make_model(graph, opset_imports=[opset_import])
        return model

    def check_models(self, interleaved: bool, model_type: str):
        initializers = self.create_initializers()

        expected_model_filename = "expected_model.onnx"
        expected_model = self.create_fused_model(interleaved, initializers)
        onnx.save(expected_model, expected_model_filename)

        original_model_filename = "original_model.onnx"
        use_redundant_squeeze_ops = True
        original_model = self.create_test_model(model_type, use_redundant_squeeze_ops, initializers)
        onnx.save(original_model, original_model_filename)

        self.verify_fusion(expected_model_filename, original_model_filename)
        os.remove(original_model_filename)

        use_redundant_squeeze_ops = False
        original_model = self.create_test_model(model_type, use_redundant_squeeze_ops, initializers)
        onnx.save(original_model, original_model_filename)

        self.verify_fusion(expected_model_filename, original_model_filename)
        os.remove(expected_model_filename)
        os.remove(original_model_filename)

    # Hugging Face's `decoder_model.onnx`
    def test_hf_decoder_model(self):
        interleaved = False  # HF model does not use interleaving
        model_type = "no_past"
        self.check_models(interleaved, model_type)

    # Hugging Face's `decoder_with_past_model.onnx`
    def test_hf_decoder_with_past_model(self):
        interleaved = False  # HF model does not use interleaving
        model_type = "past"
        self.check_models(interleaved, model_type)

    # Hugging Face's `decoder_merged.onnx`
    def test_hf_decoder_merged_model(self):
        interleaved = False  # HF model does not use interleaving
        model_type = "merged"
        self.check_models(interleaved, model_type)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

# This script is a stub that uses the model conversion script from the util subdirectory.
# We do it this way so we can use relative imports in that script, which makes it easy to include
# in the ORT python package (where it must use relative imports)
from util.convert_onnx_models_to_ort import convert_onnx_models_to_ort, parse_args

if __name__ == "__main__":
    args = parse_args()
    convert_onnx_models_to_ort(
        args.model_path_or_dir,
        output_dir=args.output_dir,
        optimization_styles=args.optimization_style,
        custom_op_library_path=args.custom_op_library,
        target_platform=args.target_platform,
        save_optimized_onnx_model=args.save_optimized_onnx_model,
        allow_conversion_failures=args.allow_conversion_failures,
        enable_type_reduction=args.enable_type_reduction,
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import argparse
from dataclasses import dataclass

import numpy as np
from benchmark import BenchmarkOp, add_arguments


@dataclass
class OpParam:
    x: int
    y: int
    m: int
    n: int
    input_data_type: type
    output_data_type: type


@dataclass
class ModelParam:
    token_type_ids_dim0: int
    input_ids_dim1: int


class BenchmarkCast(BenchmarkOp):
    def __init__(self, args):
        BenchmarkOp.__init__(self, args)

    @classmethod
    def create_inputs_outputs(cls, op_param):
        np.random.seed(0)
        input_data = np.random.rand(op_param.x, op_param.y, op_param.m, op_param.n).astype(op_param.input_data_type)
        output_data = np.random.rand(op_param.x, op_param.y, op_param.m, op_param.n).astype(op_param.output_data_type)
        inputs = {"X": input_data}
        outputs = {"Y": output_data}
        return inputs, outputs

    def add_model_cases(self, mp, model, input_data_type, output_data_type):
        self.add_case(
            OpParam(
                1,
                mp.token_type_ids_dim0,
                mp.input_ids_dim1,
                1024,
                input_data_type,
                output_data_type,
            ),
            model,
        )
        self.add_case(
            OpParam(
                1,
                mp.token_type_ids_dim0,
                mp.input_ids_dim1,
                1,
                input_data_type,
                output_data_type,
            ),
            model,
        )
        self.add_case(
            OpParam(
                16,
                mp.token_type_ids_dim0,
                mp.input_ids_dim1,
                mp.input_ids_dim1,
                input_data_type,
                output_data_type,
            ),
            model,
        )

    def create_cases(self):
        model = "models/cast_fp16tofp32.onnx" if self.args.precision == "fp16" else "models/cast_fp32tofp16.onnx"
        input_data_type = np.float16 if self.args.precision == "fp16" else np.float32
        output_data_type = np.float32 if self.args.precision == "fp16" else np.float16
        # huggingface bert-large
        self.add_case(OpParam(1, 1, 1, 1024, input_data_type, output_data_type), model)
        self.add_case(OpParam(1, 1, 1024, 1024, input_data_type, output_data_type), model)
        self.add_case(OpParam(1, 1, 1024, 4096, input_data_type, output_data_type), model)
        self.add_case(OpParam(1, 1, 1024, 30522, input_data_type, output_data_type), model)
        # huggingface bert-large with default dims
        model_param = ModelParam(8, 512)
        self.add_model_cases(model_param, model, input_data_type, output_data_type)
        # huggingface bert-large with large input dims
        model_param = ModelParam(32, 1024)
        self.add_model_cases(model_param, model, input_data_type, output_data_type)

    @classmethod
    def case_profile(cls, op_param, time):
        profile = f"(x y m n input_data_type) = ({op_param.x} {op_param.y} {op_param.m} {op_param.n} {op_param.input_data_type}), {time:7.4f} ms"
        return profile


def main():
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    bm = BenchmarkCast(args)
    bm.benchmark()


if __name__ == "__main__":
    main()



import unittest

import onnx
import pytest
import torch
from parity_utilities import find_transformers_source

if find_transformers_source():
    from benchmark_helper import ConfigModifier, OptimizerInfo, Precision
    from huggingface_models import MODELS
    from onnx_exporter import export_onnx_model_from_pt
    from shape_infer_helper import SymbolicShapeInferenceHelper
else:
    from onnxruntime.transformers.benchmark_helper import ConfigModifier, OptimizerInfo, Precision
    from onnxruntime.transformers.huggingface_models import MODELS
    from onnxruntime.transformers.onnx_exporter import export_onnx_model_from_pt
    from onnxruntime.transformers.shape_infer_helper import SymbolicShapeInferenceHelper


class SymbolicShapeInferenceHelperTest(unittest.TestCase):
    def _load_onnx(self, model_name):
        input_names = MODELS[model_name][0]
        base_path = "../onnx_models/"

        config_modifier = ConfigModifier(None)
        fusion_options = None
        model_class = "AutoModel"
        with torch.no_grad():
            export_onnx_model_from_pt(
                model_name,
                MODELS[model_name][1],
                MODELS[model_name][2],
                MODELS[model_name][3],
                model_class,
                config_modifier,
                "../cache_models",
                base_path,
                input_names[:1],
                False,
                Precision.FLOAT32,
                OptimizerInfo.BYSCRIPT,
                True,
                True,
                False,
                {},
                fusion_options,
            )
        model_path = base_path + model_name.replace("-", "_") + "_1.onnx"
        return onnx.load_model(model_path)

    # TODO: use a static lightweight model for test
    @pytest.mark.slow
    def test_bert_shape_infer_helper(self):
        model = self._load_onnx("bert-base-cased")
        shape_infer_helper = SymbolicShapeInferenceHelper(model)
        self.assertEqual(shape_infer_helper.infer({"batch_size": 4, "seq_len": 16}), True)
        self.assertEqual(shape_infer_helper.get_edge_shape("802"), [])
        self.assertEqual(shape_infer_helper.get_edge_shape("804"), [4, 16, 3072])
        self.assertEqual(shape_infer_helper.get_edge_shape("1748"), [1])
        self.assertEqual(
            shape_infer_helper.get_edge_shape("encoder.layer.4.attention.output.LayerNorm.weight"),
            [768],
        )
        self.assertEqual(shape_infer_helper.get_edge_shape("817"), [4, 16, 1])
        self.assertEqual(
            shape_infer_helper.get_edge_shape("encoder.layer.4.intermediate.dense.bias"),
            [3072],
        )
        self.assertEqual(shape_infer_helper.get_edge_shape("880"), [4, 12, 16, 16])

        self.assertEqual(shape_infer_helper.compare_shape("329", "253"), False)
        self.assertEqual(shape_infer_helper.compare_shape("447", "371"), False)
        self.assertEqual(shape_infer_helper.compare_shape("329", "817"), True)
        self.assertEqual(shape_infer_helper.compare_shape("447", "853"), False)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import onnx
from onnx import TensorProto, helper


# Create a model with shared initializers that can be updated in-place by the transpose optimizer,
# including ones behind a DQ node. The transpose optimizer updates the first usage and inserts
# Transpose/Unsqueeze ops on the others (see UnsqueezeInput and TransposeInput).
# When we push the Transpose past other usages we should be able to cancel out those Transpose/Unsqueeze ops.
# We need 3 DQ nodes to ensure the Transpose or Unsqueeze added by the transpose optimizer is not
# removed prematurely.
def create_model(broadcast_weights: bool):
    if broadcast_weights:
        bias_shape = [2, 2]
        bias_values = np.random.randn(2, 2)
    else:
        bias_shape = [1, 3, 2, 2]
        bias_values = np.random.randn(1, 3, 2, 2)

    graph = helper.make_graph(
        name="graph",
        inputs=[
            helper.make_tensor_value_info("input0", TensorProto.FLOAT, [1, 2, 2, 3]),
        ],
        initializer=[
            helper.make_tensor("bias_quant", TensorProto.UINT8, bias_shape, bias_values.astype(np.uint8)),
            helper.make_tensor("bias_fp32", TensorProto.FLOAT, bias_shape, bias_values.astype(np.float32)),
            helper.make_tensor("dq_scale0", TensorProto.FLOAT, [], [1.5]),
            helper.make_tensor("dq_zp0", TensorProto.UINT8, [], [5]),
            helper.make_tensor("dq_scale1", TensorProto.FLOAT, [], [0.5]),
        ],
        nodes=[
            # Transpose input from channels last to channels first
            helper.make_node("Transpose", ["input0"], ["input_T"], perm=[0, 3, 1, 2]),
            helper.make_node("DequantizeLinear", ["bias_quant", "dq_scale0", "dq_zp0"], ["DQ0"], "DQ0"),
            helper.make_node("Add", ["input_T", "DQ0"], ["A0"], "A0"),
            helper.make_node("DequantizeLinear", ["bias_quant", "dq_scale1"], ["DQ1"], "DQ1"),
            helper.make_node("Add", ["A0", "DQ1"], ["A1"], "A1"),
            helper.make_node("DequantizeLinear", ["bias_quant", "dq_scale0"], ["DQ2"], "DQ2"),
            helper.make_node("Add", ["A1", "DQ2"], ["A2"], "A2"),
            helper.make_node("Add", ["A2", "bias_fp32"], ["A3"], "A3"),
            helper.make_node("Add", ["A3", "bias_fp32"], ["A4"], "A4"),
            # NCHW to NHWC
            helper.make_node("Transpose", ["A4"], ["output0"], perm=[0, 2, 3, 1]),
        ],
        outputs=[
            helper.make_tensor_value_info("output0", TensorProto.FLOAT, [1, 2, 2, 3]),
        ],
    )

    model = helper.make_model(graph)
    onnx.checker.check_model(model, full_check=True)
    return model


def create_model_with_Where():  # noqa 'Where' is the operator name
    """
    Create a model to validate the logic to cancel out the Transpose -> Squeeze -> DQ between an updated shared
    initializer and other usage. We need to use Where as we require more than 2 inputs.
    The `condition` input will be having a Transpose pushed through it will have a negative cost.
    The `X` input will have a positive cost which cancels out the negative value.
    The `Y` input will be a shared initializer that is broadcast. If we don't find the Transpose to make the cost of it
    negative we will not push the Transpose though.

    If we only have 2 inputs, the broadcast initializer will always cost less due to its smaller rank, meaning we don't
    actually need to look for the Squeeze in that case.
    """
    cond_0_shape = [3, 2]  # transpose to 2, 3
    cond_1_shape = [2, 3]
    x_0_shape = [3]  # broadcast so Transpose goes through Where0
    x_1_shape = [3]  # also broadcast
    y_shape = [3]  # should be transposed and broadcast to [3, 1] if we push the transpose through the Where
    y_values = np.random.randn(3)

    graph = helper.make_graph(
        name="graph",
        inputs=[
            helper.make_tensor_value_info("cond_in_0", TensorProto.BOOL, cond_0_shape),
            helper.make_tensor_value_info("cond_in_1", TensorProto.BOOL, cond_1_shape),
            helper.make_tensor_value_info("x_in_0", TensorProto.FLOAT, x_0_shape),
            helper.make_tensor_value_info("x_in_1", TensorProto.FLOAT, x_1_shape),
        ],
        initializer=[
            helper.make_tensor("y_quant", TensorProto.UINT8, y_shape, y_values.astype(np.uint8)),
            helper.make_tensor("dq_scale0", TensorProto.FLOAT, [], [1.5]),
            helper.make_tensor("dq_scale1", TensorProto.FLOAT, [], [0.5]),
        ],
        nodes=[
            # Transpose the cond input
            helper.make_node("Transpose", ["cond_in_0"], ["cond_in_T"], perm=[1, 0]),
            helper.make_node("DequantizeLinear", ["y_quant", "dq_scale0"], ["DQ0"], "DQ0"),
            # first usage of shared initializer. simple so we know the Transpose can push through it
            helper.make_node("Where", ["cond_in_T", "x_in_0", "DQ0"], ["Where0"], "Where0"),
            helper.make_node("DequantizeLinear", ["y_quant", "dq_scale1"], ["DQ1"], "DQ1"),
            helper.make_node("Add", ["x_in_1", "Where0"], ["Add0"], "Add0"),
            # second usage of shared initializer. requires looking past the Squeeze to push the transpose through
            helper.make_node("Where", ["cond_in_1", "Add0", "DQ1"], ["Where1"], "Where1"),
            helper.make_node("Transpose", ["Where1"], ["output0"], perm=[1, 0]),
        ],
        outputs=[
            helper.make_tensor_value_info("output0", TensorProto.FLOAT, [3, 2]),
        ],
    )

    model = helper.make_model(graph)
    onnx.checker.check_model(model, full_check=True)
    return model


if __name__ == "__main__":
    model = create_model(broadcast_weights=False)
    onnx.save(model, "transpose_optimizer_shared_initializers.onnx")
    model = create_model(broadcast_weights=True)
    onnx.save(model, "transpose_optimizer_shared_initializers_broadcast.onnx")
    model = create_model_with_Where()
    onnx.save(model, "transpose_optimizer_shared_initializers_broadcast2.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
#
# This script evaluates accuracy of ONNX models for question-answering task on SQuAD data set.
# Example to evaluate raw and optimized model for CUDA in Linux:
#   pip3 install datasets evaluate optimum transformers onnxruntime-gpu
#
#   python3 eval_squad.py -m bert-large-uncased-whole-word-masking-finetuned-squad -s 384 -b 1 --use_io_binding
#
#   python3 -m onnxruntime.transformers.optimizer \
#           --input ./bert-large-uncased-whole-word-masking-finetuned-squad/model.onnx \
#           --output ./bert-large-uncased-whole-word-masking-finetuned-squad/optimized_model.onnx
#
#   python3 eval_squad.py -m bert-large-uncased-whole-word-masking-finetuned-squad -s 384 -b 1 --use_io_binding \
#           --onnx ./bert-large-uncased-whole-word-masking-finetuned-squad/optimized_model.onnx
#
#   Snippet of example output in A100:
#   {'exact': 86.65089877010406, 'f1': 92.99433524952254, 'total': 10570, 'HasAns_exact': 86.65089877010406
#    'total_time_in_seconds': 81.69239814393222, 'samples_per_second': 129.387804008115,
#    'latency_in_seconds': 0.007728703703304846, 'provider': 'CUDAExecutionProvider',
#    'pretrained_model_name': 'bert-large-uncased-whole-word-masking-finetuned-squad',
#    'batch_size': 1, 'sequence_length': 384, 'use_io_binding': True}
import argparse
import csv
import os
import time

try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version

from pathlib import Path
from typing import Any

from datasets import load_dataset
from evaluate import evaluator
from optimum.onnxruntime import ORTModelForQuestionAnswering
from optimum.version import __version__ as optimum_version
from packaging import version as version_check
from transformers import AutoTokenizer, pipeline

if version_check.parse(optimum_version) < version_check.parse("1.13.1"):
    raise ImportError(f"Please install optimum>=1.13.1. Current version: {optimum_version}.")

PRETRAINED_SQUAD_MODELS = [
    "bert-large-uncased-whole-word-masking-finetuned-squad",
    "deepset/roberta-base-squad2",
    "distilbert-base-cased-distilled-squad",
]


def get_package_version(package_name: str):
    try:
        return version(package_name)
    except PackageNotFoundError:
        return None


def load_onnx_model(
    model_id: str, onnx_path: str | None = None, provider="CUDAExecutionProvider", use_io_binding: bool = False
):
    """Load onnx model given pretrained model name and optional ONNX model path. If onnx_path is None,
    the default onnx model from optimum will be used.

    Args:
        model_id (str): pretrained model name or checkpoint path
        onnx_path (Optional[str], optional): path of onnx model to evaluate. Defaults to None.

    Returns:
        model: ORTModel for the onnx model
        onnx_path: the path of onnx model
    """

    if onnx_path is None:
        # Export onnx to a sub-directory named by the model id
        model = ORTModelForQuestionAnswering.from_pretrained(
            model_id, export=True, provider=provider, use_io_binding=use_io_binding
        )
        save_onnx_dir = os.path.join(".", model_id)
        model.save_pretrained(save_onnx_dir)
        onnx_path = os.path.join(save_onnx_dir, "model.onnx")
        print("Model is exported to onnx file:", onnx_path)
    else:
        model = ORTModelForQuestionAnswering.from_pretrained(
            os.path.dirname(onnx_path),
            file_name=Path(onnx_path).name,
            provider=provider,
            use_io_binding=use_io_binding,
            # provider_options={"enable_skip_layer_norm_strict_mode": True},
        )

    return model, onnx_path


def output_details(results: list[dict[str, Any]], csv_filename: str):
    """Output a CSV file with detail of each test results.

    Args:
        results (List[Dict[str, Any]]): list of JSON results.
        csv_filename (str): path of output CSV file
    """
    with open(csv_filename, mode="a", newline="", encoding="ascii") as csv_file:
        column_names = [
            "pretrained_model_name",
            "onnx_path",
            "provider",
            "disable_fused_attention",
            "batch_size",
            "sequence_length",
            "use_io_binding",
            "exact",
            "f1",
            "total",
            "HasAns_exact",
            "HasAns_f1",
            "HasAns_total",
            "best_exact",
            "best_exact_thresh",
            "best_f1",
            "best_f1_thresh",
            "total_time_in_seconds",
            "samples_per_second",
            "latency_in_seconds",
        ]

        csv_writer = csv.DictWriter(csv_file, fieldnames=column_names)
        csv_writer.writeheader()
        for result in results:
            csv_writer.writerow(result)

        csv_file.flush()

    print(f"Detail results are saved to csv file: {csv_filename}")


def output_summary(results: list[dict[str, Any]], csv_filename: str, metric_name: str):
    """Output a CSV file with summary of a metric on combinations of batch_size and sequence_length.

    Args:
        results (List[Dict[str, Any]]): list of JSON results.
        csv_filename (str): path of output CSV file
        metric_name (str): the metric to summarize
    """
    with open(csv_filename, mode="a", newline="", encoding="ascii") as csv_file:
        header_names = [
            "pretrained_model_name",
            "onnx_path",
            "provider",
            "disable_fused_attention",
            "use_io_binding",
        ]

        model_list = list({result["onnx_path"] for result in results})
        model_list.sort()

        batch_sizes = list({result["batch_size"] for result in results})
        batch_sizes.sort()

        sequence_lengths = list({result["sequence_length"] for result in results})
        sequence_lengths.sort()

        key_names = []
        for sequence_length in sequence_lengths:
            for batch_size in batch_sizes:
                key_names.append(f"b{batch_size}_s{sequence_length}")

        csv_writer = csv.DictWriter(csv_file, fieldnames=header_names + key_names)
        csv_writer.writeheader()

        for model in model_list:
            row = {}

            # Metric value for given pair of batch_size and sequence_length.
            # Assume that (onnx_path, batch_size and sequence_length) are unique so keep first occurrence only.
            values = {}
            values.update({k: "" for k in key_names})

            for result in results:
                if result["onnx_path"] == model and result[metric_name]:
                    headers = {k: v for k, v in result.items() if k in header_names}
                    if not row:
                        row.update(headers)

                    batch_size = result["batch_size"]
                    sequence_length = result["sequence_length"]
                    key = f"b{batch_size}_s{sequence_length}"

                    if key in key_names:
                        values[key] = result[metric_name]

            if row:
                for key in key_names:
                    row[key] = values.get(key, "")
                csv_writer.writerow(row)

        csv_file.flush()

    print(f"Summary results for {metric_name} are saved to csv file: {csv_filename}")


def main():
    args = parse_arguments()
    print(args)

    for name in ["onnxruntime-gpu", "onnxruntime", "onnx", "torch", "transformers", "optimum", "datasets", "evaluate"]:
        package_version = get_package_version(name)
        if package_version:
            print(f"{name} version", package_version)

    pretrained_model_name = args.model_name
    if args.onnx and not os.path.exists(args.onnx):
        raise RuntimeError(f"Onnx model path does not exist: {args.onnx}")

    disable_fused_attention = os.environ.get("ORT_DISABLE_FUSED_ATTENTION", "0") == "1"

    all_results = []
    tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name)
    for sequence_length in args.sequence_lengths:
        tokenizer.model_max_length = sequence_length
        tokenizer.doc_stride = min(sequence_length // 2, 128)
        if args.onnx is None:
            print("Exporting onnx model. It might take a few minutes...")
        start_time = time.time()
        ort_model, onnx_path = load_onnx_model(pretrained_model_name, args.onnx, args.provider, args.use_io_binding)
        latency = time.time() - start_time
        print(f"Onnx model exported or loaded in {latency:.1f} seconds")

        print(ort_model.config)
        if sequence_length > ort_model.config.max_position_embeddings:
            raise RuntimeError("sequence length should not be larger than {ort_model.config.max_position_embeddings}")

        qa_pipeline = pipeline(
            "question-answering", model=ort_model, tokenizer=tokenizer, question_first=True, batch_size=args.batch_size
        )

        task_evaluator = evaluator("question-answering")
        print("Loading dataset...")
        start_time = time.time()
        squad_dataset = load_dataset("squad", split=f"validation[:{args.total}]" if args.total > 0 else "validation")
        latency = time.time() - start_time
        print(f"Dataset loaded in {latency:.1f} seconds")

        print("Evaluating squad_v2 with ORT. It might take a few minutes...")
        start_time = time.time()
        result = task_evaluator.compute(
            model_or_pipeline=qa_pipeline,
            data=squad_dataset,
            metric="squad_v2",
            squad_v2_format=True,
        )
        latency = time.time() - start_time
        print(f"Evaluation done in {latency:.1f} seconds")

        result["provider"] = args.provider
        result["disable_fused_attention"] = disable_fused_attention
        result["pretrained_model_name"] = pretrained_model_name
        result["onnx_path"] = onnx_path
        result["batch_size"] = args.batch_size
        result["sequence_length"] = sequence_length
        result["use_io_binding"] = args.use_io_binding
        print(result)

        all_results.append(result)

    output_details(all_results, "detail.csv")

    for metric_name in ["f1", "exact", "samples_per_second"]:
        output_summary(all_results, f"{metric_name}.csv", metric_name)


def parse_arguments(argv=None):
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-m",
        "--model_name",
        required=False,
        type=str,
        default=PRETRAINED_SQUAD_MODELS[0],
        help=f"Checkpoint directory or pre-trained model names in the list: {PRETRAINED_SQUAD_MODELS}",
    )

    parser.add_argument(
        "-s",
        "--sequence_lengths",
        nargs="+",
        type=int,
        default=[384],
        help="Sequence lengths for onnx model inputs. It could have multiple values.",
    )

    parser.add_argument(
        "-b",
        "--batch_size",
        type=int,
        default=1,
        help="batch size for inference.",
    )

    parser.add_argument("-t", "--total", type=int, default=0, help="Total samples to test. 0 means all samples.")

    parser.add_argument(
        "--onnx",
        required=False,
        type=str,
        default=None,
        help="Optional onnx model path. If not specified, optimum will be used to export onnx model for testing.",
    )

    parser.add_argument(
        "--provider",
        required=False,
        default="CUDAExecutionProvider",
        help="Select which Execution Provider to use for runs. Default is CUDAExecutionProvider.",
    )

    parser.add_argument("--use_io_binding", required=False, action="store_true", help="Use IO Binding for GPU.")
    parser.set_defaults(use_io_binding=False)

    args = parser.parse_args(argv)

    return args


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
import csv
import datetime
import os
import re

import pandas as pd
from azure.kusto.data import KustoConnectionStringBuilder
from azure.kusto.ingest import QueuedIngestClient
from post import get_identifier, parse_arguments, write_table


def parse_valgrind_log(input_path, output_path, keywords):
    is_definitely_lost = False
    is_ort_trt_related = False
    buffer = []
    leak_block = None
    leak_bytes = None
    keyword = None
    results = []

    with open(input_path) as file:
        lines = file.readlines()

        for line in lines:
            line = line.strip()  # noqa: PLW2901
            # Remove "==xxxxx==" pattern from the line
            line = line.split("==")[-1].strip()  # noqa: PLW2901

            if "blocks are definitely lost in loss" in line:
                is_definitely_lost = True
                # Extract LeakBlock and LeakBytes
                match = re.search(r"([\d,]+) byte[s]? in ([\d,]+) block[s]?", line)
                if match:
                    leak_bytes = match.group(1).replace(",", "")
                    leak_block = match.group(2).replace(",", "")
                continue

            if is_definitely_lost:
                if line:
                    buffer.append(line)
                    for word in keywords:
                        if word in line:
                            is_ort_trt_related = True
                            keyword = word
                            break

            # End of section
            if is_definitely_lost and not line:
                if is_ort_trt_related:
                    results.append((keyword, leak_block, leak_bytes, "\n".join(buffer)))
                # Reset var
                is_definitely_lost = False
                is_ort_trt_related = False
                buffer = []
                leak_block = None
                leak_bytes = None
                keyword = None

    # Writing results to CSV
    with open(output_path, "w", newline="") as csvfile:
        csvwriter = csv.writer(csvfile)
        csvwriter.writerow(["Keyword", "LeakBlock", "LeakBytes", "ValgrindMessage"])
        for entry in results:
            csvwriter.writerow([entry[0], entry[1], entry[2], entry[3]])


def parse_concurrency_test_log(input_path, output_path):
    with open(input_path) as log_file:
        log_content = log_file.read()

    failed_cases_section = log_content.split("Failed Test Cases:")[1]

    # passed = 1 if no failed test cases
    if failed_cases_section.strip() == "":
        passed = 1
    else:
        passed = 0

    with open(output_path, "w", newline="") as csv_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerow(["Passed", "Log"])
        csv_writer.writerow([passed, log_content])


if __name__ == "__main__":
    args = parse_arguments()

    # connect to database
    kcsb_ingest = KustoConnectionStringBuilder.with_az_cli_authentication(args.kusto_conn)
    ingest_client = QueuedIngestClient(kcsb_ingest)
    identifier = get_identifier(
        args.commit_datetime, args.commit_hash, args.trt_version, args.branch, args.use_tensorrt_oss_parser
    )
    upload_time = datetime.datetime.now(tz=datetime.timezone.utc).replace(microsecond=0)

    try:
        result_mem_test_path = args.report_folder
        os.chdir(result_mem_test_path)
        # Parse mem_test log
        logs = ["valgrind.log", "concurrency_test.log"]
        csv_paths = ["mem_test.csv", "concurrency_test.csv"]
        for log, csv_path in zip(logs, csv_paths, strict=False):
            if os.path.exists(log):
                print(f"{identifier}: Parsing {log}")
                if log == logs[0]:
                    parse_valgrind_log(log, csv_path, ["TensorrtExecutionProvider", "TensorRT"])
                else:
                    parse_concurrency_test_log(log, csv_path)

        # Upload to db
        for csv_path, db_table_name in zip(
            csv_paths, ["ep_valgrind_record", "ep_concurrencytest_record"], strict=False
        ):
            if os.path.exists(csv_path):
                table = pd.read_csv(csv_path)
                write_table(
                    ingest_client,
                    args.database,
                    table,
                    db_table_name,
                    upload_time,
                    identifier,
                    args.branch,
                    args.commit_hash,
                    args.commit_datetime,
                )
                print(f"{identifier}: {csv_path} is synced to db")

    except Exception as e:
        print(str(e))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# _fallback_exceptions.py


class ORTModuleFallbackException(Exception):  # noqa: N818
    """Base exception class for fallback

    Although it must be specialized for specific scenarios,
    it can also be used for generic exception that require fallback
    """


class ORTModuleInitException(ORTModuleFallbackException):
    """Trigger fallback for ORTModule initialization related exceptions

    This exception is triggered when an incompatible or missing requirements for ORTModule are detected,
    including PyTorch version, missing ORTModule's PyTorch C++ extension binaries, etc.
    """


class ORTModuleDeviceException(ORTModuleFallbackException):
    """Trigger fallback for device related exceptions

    NOTE: This exception is raised during device validation within ORTModule frontend.
    Some device related exceptions can only be detected during PyTorch ONNX exporter execution.
    This exception does not capture these scenarios.
    """


class ORTModuleIOError(ORTModuleFallbackException):
    """Trigger fallback for I/O related exceptions

    NOTE: This exception is raised during I/O validation within ORTModule Frontend.
    Some I/O related exceptions can only be detected during PyTorch ONNX exporter execution.
    This exception does not capture these scenarios.
    """


class ORTModuleTorchModelException(ORTModuleFallbackException):
    """Trigger fallback for PyTorch modules related exceptions

    This exception is raised during model validation within ORTModule frontend and is based on
    checking type(model) over a hardcoded list of incompatible models.
    """


class ORTModuleONNXModelException(ORTModuleFallbackException):
    """Trigger fallback for ONNX model related exceptions

    This exception is raised during model conversion to ONNX and post-processing validation within ORTModule frontend.
    """


def wrap_exception(
    new_exception: ORTModuleFallbackException, raised_exception: Exception
) -> ORTModuleFallbackException:
    """Wraps `raised_exception` exception as cause for the returned `new_exception` exception"""

    exception = None
    try:
        raise new_exception(raised_exception) from raised_exception
    except Exception as e:
        exception = e
    return exception



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
#
# This script generates test code coverage for Android.
# The prerequistes:
#     1. The Onnxruntime build with coverage option to compile/link the source files using --coverage optoin
#     2. The tests are run on the target emulator and *.gcda files are available on the emulator
#     3. The emulator which ran tests must be running. Otherwise this script will fail

import argparse
import os
import sys

from build import run_subprocess

SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
REPO_DIR = os.path.normpath(os.path.join(SCRIPT_DIR, "..", ".."))

sys.path.append(os.path.join(REPO_DIR, "tools", "python"))

import util.android as android  # noqa: E402


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--build_dir", required=True, help="Path to the build directory.")
    parser.add_argument(
        "--config",
        default="Debug",
        choices=["Debug", "MinSizeRel", "Release", "RelWithDebInfo"],
        help="Configuration(s) to run code coverage.",
    )
    parser.add_argument("--android_sdk_path", required=True, help="The Android SDK root.")
    return parser.parse_args()


def main():
    args = parse_arguments()

    sdk_tool_paths = android.get_sdk_tool_paths(args.android_sdk_path)

    def adb_pull(src, dest, **kwargs):
        return run_subprocess([sdk_tool_paths.adb, "pull", src, dest], **kwargs)

    def adb_shell(*args, **kwargs):
        return run_subprocess([sdk_tool_paths.adb, "shell", *args], **kwargs)

    script_dir = os.path.realpath(os.path.dirname(__file__))
    source_dir = os.path.normpath(os.path.join(script_dir, "..", ".."))
    cwd = os.path.abspath(os.path.join(args.build_dir, args.config))
    adb_shell("cd /data/local/tmp && tar -zcf gcda_files.tar.gz *.dir")
    adb_pull("/data/local/tmp/gcda_files.tar.gz", cwd)
    os.chdir(cwd)
    run_subprocess("tar -zxf gcda_files.tar.gz -C CMakeFiles".split(" "))
    cmd = ["gcovr", "-s", "-r"]
    cmd.append(os.path.join(source_dir, "onnxruntime"))
    cmd.extend([".", "-o"])
    cmd.append(os.path.join(cwd, "coverage_rpt.txt"))
    run_subprocess(cmd, cwd=os.path.join(cwd, "CMakeFiles"))


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------


import torch

# Mapping from pytorch scalar type to onnx scalar type.
_CAST_PYTORCH_TO_ONNX = {
    "Byte": [torch.onnx.TensorProtoDataType.UINT8, torch.uint8],
    "Char": [torch.onnx.TensorProtoDataType.INT8, torch.int8],
    "Double": [torch.onnx.TensorProtoDataType.DOUBLE, torch.double],
    "Float": [torch.onnx.TensorProtoDataType.FLOAT, torch.float],
    "Half": [torch.onnx.TensorProtoDataType.FLOAT16, torch.half],
    "Int": [torch.onnx.TensorProtoDataType.INT32, torch.int],
    "Long": [torch.onnx.TensorProtoDataType.INT64, torch.int64],
    "Short": [torch.onnx.TensorProtoDataType.INT16, torch.short],
    "Bool": [torch.onnx.TensorProtoDataType.BOOL, torch.bool],
    "ComplexFloat": [torch.onnx.TensorProtoDataType.COMPLEX64, torch.complex64],
    "ComplexDouble": [torch.onnx.TensorProtoDataType.COMPLEX128, torch.complex128],
    "BFloat16": [torch.onnx.TensorProtoDataType.BFLOAT16, torch.bfloat16],
    # Not yet defined in torch.
    # "Float8E4M3FN": torch.onnx.TensorProtoDataType.FLOAT8E4M3FN,
    # "Float8E4M3FNUZ": torch.onnx.TensorProtoDataType.FLOAT8E4M3FNUZ,
    # "Float8E5M2": torch.onnx.TensorProtoDataType.FLOAT8E5M2,
    # "Float8E5M2FNUZ": torch.onnx.TensorProtoDataType.FLOAT8E5M2FNUZ,
    "Undefined": [torch.onnx.TensorProtoDataType.UNDEFINED, None],
}


_DTYPE_TO_ONNX = {torch_dtype: onnx_dtype for k, (onnx_dtype, torch_dtype) in _CAST_PYTORCH_TO_ONNX.items()}

_ONNX_TO_DTYPE = {onnx_dtype: torch_dtype for torch_dtype, onnx_dtype in _DTYPE_TO_ONNX.items()}


def pytorch_type_to_onnx_dtype(dtype_or_scalar_type: torch.dtype | str) -> torch.onnx.TensorProtoDataType:
    """Converts a pytorch dtype or scalar type string to an onnx dtype.
    PyTorch type can be either a dtype or a scalar type string.
    """
    dtype = dtype_or_scalar_type
    if isinstance(dtype, str):
        if dtype not in _CAST_PYTORCH_TO_ONNX:
            raise RuntimeError(f"Unsupported dtype {dtype}")
        return _CAST_PYTORCH_TO_ONNX[dtype][0]

    if dtype not in _DTYPE_TO_ONNX:
        raise RuntimeError(f"Unsupported dtype {dtype}")
    return _DTYPE_TO_ONNX[dtype]


def pytorch_scalar_type_to_pytorch_dtype(dtype: str) -> torch.dtype:
    """Converts a pytorch scalar type string to a pytorch dtype."""
    assert isinstance(dtype, str)
    if dtype not in _CAST_PYTORCH_TO_ONNX:
        raise RuntimeError(f"Unsupported dtype {dtype}")
    return _CAST_PYTORCH_TO_ONNX[dtype][1]


def onnx_dtype_to_pytorch_dtype(dtype: torch.onnx.TensorProtoDataType) -> torch.dtype:
    """Converts an onnx dtype to a pytorch dtype."""
    if dtype not in _ONNX_TO_DTYPE:
        raise RuntimeError(f"Unsupported dtype {dtype}")
    return _ONNX_TO_DTYPE[dtype]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import copy
import itertools

import onnx
import sympy
from onnx import GraphProto, ModelProto, NodeProto, TensorProto, helper

from ._common import SymbolicDSU, TensorInfo, TypeAndShapeInfer
from ._decompose import DecomposeDispatch
from ._op_config import is_elementwise_node
from ._sympy_utils import parse_shape
from ._utils import get_attribute, to_torch_tensor, topological_sort


class SortedGraph:
    """
    This class is used to
        1. decompose complex operators into preliminary operators,
        2. sort the operators in topological order,
        3. infer the type and shape of each node inputs and outputs.

    input args:
        model: the ONNX model.
        input_shapes: the shapes of the model inputs. Can be numeric values or symbolic values.
    """

    def __init__(self, model: ModelProto, input_shapes: list[list[sympy.Expr]]):
        self._model: ModelProto = model
        self._graph: GraphProto = model.graph
        self._input_shapes: list[list[sympy.Expr]] = input_shapes

        # For elementwise graph outputs, when we group nodes to different kernels, if the target shape is different
        # from other nodes' target shape, even it can be broadcasted, we still need to create a new kernel for it.
        self._elementwise_graph_outputs: set[str] = set()
        graph_output_names = [output.name for output in self._graph.output]
        for node in self._graph.node:
            if is_elementwise_node(node):
                self._elementwise_graph_outputs.update(
                    [output for output in node.output if output in graph_output_names]
                )

        # Topological sort the nodes in the graph.
        self._sorted_nodes: list[NodeProto] = topological_sort(
            [input.name for input in self._graph.input] + [initializer.name for initializer in self._graph.initializer],
            self._graph.node,
        )

        self._node_arg_infos: dict[str, TensorInfo] = {}
        for idx, input in enumerate(self._graph.input):
            self._node_arg_infos[input.name] = TensorInfo(input.type.tensor_type.elem_type, self._input_shapes[idx])
        for initializer in self._graph.initializer:
            self._node_arg_infos[initializer.name] = TensorInfo(
                initializer.data_type,
                parse_shape(list(initializer.dims)),
            )

        # Decompose complex operators.
        self._decompose()

        # Sort the initializers in reference order.
        # We try to reuse Triton module for different ONNX models with same graph structure,
        # even the node args names in the models are different.
        # Sorting the initializers can help to generate same model key for different ONNX models.
        initializers = {}
        for initializer in self._graph.initializer:
            initializers[initializer.name] = initializer
        self._sorted_initializers: list[TensorProto] = []
        for node in self._sorted_nodes:
            for input in node.input:
                if input in initializers:
                    self._sorted_initializers.append(initializers[input])
                    initializers.pop(input)

        # Split nodes to constant nodes and non-constant nodes.
        self._const_nodes: list[NodeProto] = [node for node in self._sorted_nodes if node.op_type == "Constant"]
        self._sorted_nodes: list[NodeProto] = [node for node in self._sorted_nodes if node.op_type != "Constant"]

    def __str__(self):
        """
        Generate a unique key for the model based on the graph structure, ignoring the node args names.
        We try to reuse Triton module for different ONNX models with same graph structure.
        """
        graph_inputs = []
        name_map = {}
        for idx, input in enumerate(self._graph.input):
            shape_str = str(self._input_shapes[idx]).replace(" ", "")
            graph_inputs.append(f"({input.type.tensor_type.elem_type!s},{shape_str})")
            name_map[input.name] = f"i{idx}"
        graph_inputs_str = ",".join(graph_inputs)

        constants = []
        for idx, initializer in enumerate(self._sorted_initializers):
            data_str = str(to_torch_tensor(initializer).tolist()).replace("\n", "").replace(" ", "")
            constants.append(f"({initializer.data_type},{data_str})")
            name_map[initializer.name] = f"c{idx}"

        for idx, node in enumerate(self._const_nodes):
            value_attr = get_attribute(node, "value")
            data_str = str(to_torch_tensor(value_attr).tolist()).replace("\n", "").replace(" ", "")
            constants.append(f"({value_attr.data_type},{data_str})")
            name_map[node.output[0]] = f"c{idx + len(self._sorted_initializers)}"
        constants_str = ",".join(constants)

        for idx, output in enumerate(self._graph.output):
            name_map[output.name] = f"o{idx}"

        nodes = []
        for node_idx, node in enumerate(self._sorted_nodes):
            inputs = []
            for input in node.input:
                inputs.append(name_map.get(input, input))
            inputs_str = ",".join(inputs)
            outputs = []
            for idx, output in enumerate(node.output):
                if output in name_map:
                    outputs.append(name_map[output])
                else:
                    name_map[output] = f"t{node_idx}_{idx}"
                    outputs.append(name_map[output])
            outputs_str = ",".join(outputs)
            attributes = []
            for attr in node.attribute:
                fields = [str(f[1]) for f in attr.ListFields()]
                attributes.append(f"{fields[0]}:{fields[2]}={fields[1]}")
            attributes_str = ",".join(attributes)
            nodes.append(f"{node.op_type}[{attributes_str}]({inputs_str})->({outputs_str})")
        nodes_str = ",".join(nodes)
        return f"{graph_inputs_str}|{len(self._graph.output)!s}|{constants_str}|{nodes_str}"

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    @property
    def const_nodes(self) -> list[NodeProto]:
        return self._const_nodes

    @property
    def sorted_nodes(self) -> list[NodeProto]:
        return self._sorted_nodes

    @property
    def original_graph(self) -> GraphProto:
        return self._graph

    @property
    def node_arg_infos(self) -> dict[str, TensorInfo]:
        return self._node_arg_infos

    @property
    def elementwise_graph_outputs(self) -> set[str]:
        return self._elementwise_graph_outputs

    def _decompose(self):
        dispatch = DecomposeDispatch()
        symbolics: SymbolicDSU = SymbolicDSU()
        pos = 0
        # If a node is complex, decompose it and insert the decomposed nodes at the same position.
        # All complex Ops are defined in DecomposeDispatch.
        # It's possible that the decomposed nodes are also complex, so we need to do the decompose recursively.
        # For example, decomposed nodes for "Softmax" contains "ReduceMean",
        # which will be decomposed to "ReduceSum" and "Div" further.
        while pos < len(self._sorted_nodes):
            node = self._sorted_nodes[pos]
            if node in dispatch:
                new_nodes = dispatch(node, self._graph, node_arg_infos=self._node_arg_infos)
                if len(new_nodes) != 1 or new_nodes[0] != node:
                    new_nodes = topological_sort(node.input, new_nodes)
                    self._sorted_nodes[pos : pos + 1] = new_nodes
                    continue
            if node.op_type == "Constant":
                value_attr = get_attribute(node, "value")
                self._node_arg_infos[node.output[0]] = TensorInfo(
                    value_attr.data_type,
                    parse_shape(list(value_attr.dims)),
                )
            else:
                input_infos = []
                for input in node.input:
                    input_infos.append(self._node_arg_infos[input])
                output_infos = TypeAndShapeInfer.infer(node, input_infos, self._graph, symbolics)
                for idx, output in enumerate(node.output):
                    self._node_arg_infos[output] = output_infos[idx]
            pos += 1
        for tensor_info in self._node_arg_infos.values():
            tensor_info.update_shape(symbolics)

    # Save the ONNX graphs for debug purpose. The original ONNX graph is the subgraph from backend.
    # The processed ONNX graph is the subgraph after decompose, it also contains the concrete shapes for each arg.
    def save_onnx(self, file_path_prefix):
        onnx.save(self._model, file_path_prefix + "_original.onnx")
        processed_model = copy.deepcopy(self._model)
        processed_model.graph.ClearField("node")
        processed_model.graph.node.extend(self.const_nodes)
        processed_model.graph.node.extend(self.sorted_nodes)
        for node in itertools.chain(processed_model.graph.input, processed_model.graph.output):
            node.type.tensor_type.shape.Clear()
            for dim in self.node_arg_infos[node.name].shape:
                if dim.is_number:
                    node.type.tensor_type.shape.dim.add().dim_value = int(dim)
                else:
                    node.type.tensor_type.shape.dim.add().dim_param = str(dim)
        value_infos = []
        for node in itertools.chain(self.const_nodes, self.sorted_nodes):
            for output in node.output:
                tensor_info = self.node_arg_infos[output]
                value_infos.append(
                    helper.make_tensor_value_info(
                        output,
                        tensor_info.dtype,
                        [int(dim) if dim.is_number else str(dim) for dim in tensor_info.shape],
                    )
                )
        processed_model.graph.ClearField("value_info")
        processed_model.graph.value_info.extend(value_infos)
        onnx.save(processed_model, file_path_prefix + "_processed.onnx")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import os
import re
import sys
from pathlib import Path


# What does the names of our C API tarball/zip files looks like
# os: win, linux, osx
# ep: cuda, tensorrt, None
def get_package_name(os, cpu_arch, ep, is_training_package):
    pkg_name = "onnxruntime-training" if is_training_package else "onnxruntime"
    if os == "win":
        pkg_name += "-win-"
        pkg_name += cpu_arch
        if ep == "cuda":
            pkg_name += "-cuda"
        elif ep == "tensorrt":
            pkg_name += "-tensorrt"
        elif ep == "rocm":
            pkg_name += "-rocm"
    elif os == "linux":
        pkg_name += "-linux-"
        pkg_name += cpu_arch
        if ep == "cuda":
            pkg_name += "-cuda"
        elif ep == "tensorrt":
            pkg_name += "-tensorrt"
        elif ep == "rocm":
            pkg_name += "-rocm"
    elif os == "osx":
        pkg_name = "onnxruntime-osx-" + cpu_arch
    return pkg_name


# Currently we take onnxruntime_providers_cuda from CUDA build
# And onnxruntime, onnxruntime_providers_shared and
# onnxruntime_providers_tensorrt from tensorrt build
# cuda binaries are split out into the platform dependent packages Microsoft.ML.OnnxRuntime.{Linux|Windows}
# and not included in the base Microsoft.ML.OnnxRuntime.Gpu package
def is_this_file_needed(ep, filename, package_name):
    if package_name == "Microsoft.ML.OnnxRuntime.Gpu":
        return False
    return (ep != "cuda" or "cuda" in filename) and (ep != "tensorrt" or "cuda" not in filename)


# nuget_artifacts_dir: the directory with uncompressed C API tarball/zip files
# ep: cuda, tensorrt, None
# files_list: a list of xml string pieces to append
# This function has no return value. It updates files_list directly
def generate_file_list_for_ep(nuget_artifacts_dir, ep, files_list, include_pdbs, is_training_package, package_name):
    for child in nuget_artifacts_dir.iterdir():
        if not child.is_dir():
            continue

        for cpu_arch in ["x86", "x64", "arm", "arm64"]:
            if child.name == get_package_name("win", cpu_arch, ep, is_training_package):
                child = child / "lib"  # noqa: PLW2901
                for child_file in child.iterdir():
                    suffixes = [".dll", ".lib", ".pdb"] if include_pdbs else [".dll", ".lib"]
                    if (
                        child_file.suffix in suffixes
                        and is_this_file_needed(ep, child_file.name, package_name)
                        and package_name != "Microsoft.ML.OnnxRuntime.Gpu.Linux"
                    ):
                        files_list.append(
                            '<file src="' + str(child_file) + f'" target="runtimes/win-{cpu_arch}/native"/>'
                        )
        for cpu_arch in ["x86_64", "arm64"]:
            if child.name == get_package_name("osx", cpu_arch, ep, is_training_package):
                child = child / "lib"  # noqa: PLW2901
                if cpu_arch == "x86_64":
                    cpu_arch = "x64"  # noqa: PLW2901
                for child_file in child.iterdir():
                    # Check if the file has digits like onnxruntime.1.8.0.dylib. We can skip such things
                    is_versioned_dylib = re.match(r".*[\.\d+]+\.dylib$", child_file.name)
                    if child_file.is_file() and child_file.suffix == ".dylib" and not is_versioned_dylib:
                        files_list.append(
                            '<file src="' + str(child_file) + f'" target="runtimes/osx-{cpu_arch}/native"/>'
                        )
        for cpu_arch in ["x64", "aarch64"]:
            if child.name == get_package_name("linux", cpu_arch, ep, is_training_package):
                child = child / "lib"  # noqa: PLW2901
                if cpu_arch == "x86_64":
                    cpu_arch = "x64"  # noqa: PLW2901
                elif cpu_arch == "aarch64":
                    cpu_arch = "arm64"  # noqa: PLW2901
                for child_file in child.iterdir():
                    if not child_file.is_file():
                        continue
                    if (
                        child_file.suffix == ".so"
                        and is_this_file_needed(ep, child_file.name, package_name)
                        and package_name != "Microsoft.ML.OnnxRuntime.Gpu.Windows"
                    ):
                        files_list.append(
                            '<file src="' + str(child_file) + f'" target="runtimes/linux-{cpu_arch}/native"/>'
                        )

        if child.name == "onnxruntime-android" or child.name == "onnxruntime-training-android":
            for child_file in child.iterdir():
                if child_file.suffix in [".aar"]:
                    files_list.append('<file src="' + str(child_file) + '" target="runtimes/android/native"/>')

        if child.name == "onnxruntime-ios":
            for child_file in child.iterdir():
                if child_file.suffix in [".zip"]:
                    files_list.append('<file src="' + str(child_file) + '" target="runtimes/ios/native"/>')


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="ONNX Runtime create nuget spec script (for hosting native shared library artifacts)",
        usage="",
    )
    # Main arguments
    parser.add_argument("--package_name", required=True, help="ORT package name. Eg: Microsoft.ML.OnnxRuntime.Gpu")
    parser.add_argument("--package_version", required=True, help="ORT package version. Eg: 1.0.0")
    parser.add_argument("--target_architecture", required=True, help="Eg: x64")
    parser.add_argument("--build_config", required=True, help="Eg: RelWithDebInfo")
    parser.add_argument("--ort_build_path", required=True, help="ORT build directory.")
    parser.add_argument("--native_build_path", required=True, help="Native build output directory.")
    parser.add_argument("--packages_path", required=True, help="Nuget packages output directory.")
    parser.add_argument("--sources_path", required=True, help="OnnxRuntime source code root.")
    parser.add_argument("--commit_id", required=True, help="The last commit id included in this package.")
    parser.add_argument(
        "--is_release_build",
        required=False,
        default=None,
        type=str,
        help="Flag indicating if the build is a release build. Accepted values: true/false.",
    )
    parser.add_argument(
        "--execution_provider",
        required=False,
        default="None",
        type=str,
        choices=["cuda", "dnnl", "openvino", "tensorrt", "snpe", "qnn", "None"],
        help="The selected execution provider for this build.",
    )
    parser.add_argument("--sdk_info", required=False, default="", type=str, help="dependency SDK information.")
    parser.add_argument(
        "--nuspec_name", required=False, default="NativeNuget.nuspec", type=str, help="nuget spec name."
    )

    return parser.parse_args()


def generate_id(line_list, package_name):
    line_list.append("<id>" + package_name + "</id>")


def generate_version(line_list, package_version):
    line_list.append("<version>" + package_version + "</version>")


def generate_authors(line_list, authors):
    line_list.append("<authors>" + authors + "</authors>")


def generate_owners(line_list, owners):
    line_list.append("<owners>" + owners + "</owners>")


def generate_description(line_list, package_name):
    description = ""

    if package_name == "Microsoft.AI.MachineLearning":
        description = "This package contains Windows ML binaries."
    elif "Microsoft.ML.OnnxRuntime.Training" in package_name:  # This is a Microsoft.ML.OnnxRuntime.Training.* package
        description = (
            "The onnxruntime-training native shared library artifacts are designed to efficiently train and infer "
            + "a wide range of ONNX models on edge devices, such as client machines, gaming consoles, and other "
            + "portable devices with a focus on minimizing resource usage and maximizing accuracy."
            + "See https://github.com/microsoft/onnxruntime-training-examples/tree/master/on_device_training for "
            + "more details."
        )
    elif "Microsoft.ML.OnnxRuntime.Gpu.Linux" in package_name:
        description = "This package contains Linux native shared library artifacts for ONNX Runtime with CUDA."
    elif "Microsoft.ML.OnnxRuntime.Gpu.Windows" in package_name:
        description = "This package contains Windows native shared library artifacts for ONNX Runtime with CUDA."
    elif "Intel.ML.OnnxRuntime" in package_name:
        description = "This package contains native shared library artifacts for ONNX Runtime with OpenVINO."
    elif "Microsoft.ML.OnnxRuntime" in package_name:  # This is a Microsoft.ML.OnnxRuntime.* package
        description = (
            "This package contains native shared library artifacts for all supported platforms of ONNX Runtime."
        )
    line_list.append("<description>" + description + "</description>")


def generate_copyright(line_list, copyright):
    line_list.append("<copyright>" + copyright + "</copyright>")


def generate_tags(line_list, tags):
    line_list.append("<tags>" + tags + "</tags>")


def generate_icon(line_list, icon_file):
    line_list.append("<icon>" + icon_file + "</icon>")


def generate_license(line_list):
    line_list.append('<license type="file">LICENSE</license>')


def generate_project_url(line_list, project_url):
    line_list.append("<projectUrl>" + project_url + "</projectUrl>")


def generate_repo_url(line_list, repo_url, commit_id):
    line_list.append('<repository type="git" url="' + repo_url + '"' + ' commit="' + commit_id + '" />')


def generate_readme(line_list):
    line_list.append("<readme>README.md</readme>")


def add_common_dependencies(xml_text, package_name, version):
    xml_text.append('<dependency id="Microsoft.ML.OnnxRuntime.Managed"' + ' version="' + version + '"/>')
    if package_name == "Microsoft.ML.OnnxRuntime.Gpu":
        xml_text.append('<dependency id="Microsoft.ML.OnnxRuntime.Gpu.Windows"' + ' version="' + version + '"/>')
        xml_text.append('<dependency id="Microsoft.ML.OnnxRuntime.Gpu.Linux"' + ' version="' + version + '"/>')


def generate_dependencies(xml_text, package_name, version):
    dml_dependency = '<dependency id="Microsoft.AI.DirectML" version="1.15.4"/>'

    if package_name == "Microsoft.AI.MachineLearning":
        xml_text.append("<dependencies>")
        # Support .Net Core
        xml_text.append('<group targetFramework="net5.0">')
        xml_text.append(dml_dependency)
        xml_text.append("</group>")
        # UAP10.0.16299, This is the earliest release of the OS that supports .NET Standard apps
        xml_text.append('<group targetFramework="UAP10.0.16299">')
        xml_text.append(dml_dependency)
        xml_text.append("</group>")
        # Support Native C++
        xml_text.append('<group targetFramework="native">')
        xml_text.append(dml_dependency)
        xml_text.append("</group>")
        xml_text.append("</dependencies>")
    else:
        include_dml = package_name == "Microsoft.ML.OnnxRuntime.DirectML"

        xml_text.append("<dependencies>")

        # Support .Net Core
        xml_text.append('<group targetFramework="NETCOREAPP">')
        add_common_dependencies(xml_text, package_name, version)
        if include_dml:
            xml_text.append(dml_dependency)
        xml_text.append("</group>")
        # Support .Net Standard
        xml_text.append('<group targetFramework="NETSTANDARD">')
        add_common_dependencies(xml_text, package_name, version)
        if include_dml:
            xml_text.append(dml_dependency)
        xml_text.append("</group>")
        # Support .Net Framework
        xml_text.append('<group targetFramework="NETFRAMEWORK">')
        add_common_dependencies(xml_text, package_name, version)
        if include_dml:
            xml_text.append(dml_dependency)
        xml_text.append("</group>")
        if package_name == "Microsoft.ML.OnnxRuntime":
            # Support net8.0-android
            xml_text.append('<group targetFramework="net8.0-android31.0">')
            xml_text.append('<dependency id="Microsoft.ML.OnnxRuntime.Managed"' + ' version="' + version + '"/>')
            xml_text.append("</group>")
            # Support net8.0-ios
            xml_text.append('<group targetFramework="net8.0-ios15.4">')
            xml_text.append('<dependency id="Microsoft.ML.OnnxRuntime.Managed"' + ' version="' + version + '"/>')
            xml_text.append("</group>")
            # Support net8.0-maccatalyst
            xml_text.append('<group targetFramework="net8.0-maccatalyst14.0">')
            xml_text.append('<dependency id="Microsoft.ML.OnnxRuntime.Managed"' + ' version="' + version + '"/>')
            xml_text.append("</group>")
        # Support Native C++
        if include_dml:
            xml_text.append('<group targetFramework="native">')
            xml_text.append(dml_dependency)
            xml_text.append("</group>")

        xml_text.append("</dependencies>")


def get_env_var(key):
    return os.environ.get(key)


def generate_release_notes(line_list, dependency_sdk_info):
    line_list.append("<releaseNotes>")
    line_list.append("Release Def:")

    branch = get_env_var("BUILD_SOURCEBRANCH")
    line_list.append("\t" + "Branch: " + (branch if branch is not None else ""))

    version = get_env_var("BUILD_SOURCEVERSION")
    line_list.append("\t" + "Commit: " + (version if version is not None else ""))

    build_id = get_env_var("BUILD_BUILDID")
    line_list.append(
        "\t"
        + "Build: https://aiinfra.visualstudio.com/Lotus/_build/results?buildId="
        + (build_id if build_id is not None else "")
    )

    if dependency_sdk_info:
        line_list.append("Dependency SDK: " + dependency_sdk_info)

    line_list.append("</releaseNotes>")


def generate_metadata(line_list, args):
    tags = "native ONNX Runtime ONNXRuntime Machine Learning MachineLearning"
    if "Microsoft.ML.OnnxRuntime.Training." in args.package_name:
        tags.append(" ONNXRuntime-Training Learning-on-The-Edge On-Device-Training On-Device Training")

    metadata_list = ["<metadata>"]
    generate_id(metadata_list, args.package_name)
    generate_version(metadata_list, args.package_version)
    generate_authors(metadata_list, "Microsoft")
    generate_owners(metadata_list, "Microsoft")
    generate_description(metadata_list, args.package_name)
    generate_copyright(metadata_list, "\xc2\xa9 " + "Microsoft Corporation. All rights reserved.")
    generate_tags(metadata_list, tags)
    generate_icon(metadata_list, "ORT_icon_for_light_bg.png")
    generate_license(metadata_list)
    generate_project_url(metadata_list, "https://github.com/Microsoft/onnxruntime")
    generate_repo_url(metadata_list, "https://github.com/Microsoft/onnxruntime.git", args.commit_id)
    generate_readme(metadata_list)
    generate_dependencies(metadata_list, args.package_name, args.package_version)
    generate_release_notes(metadata_list, args.sdk_info)
    metadata_list.append("</metadata>")

    line_list += metadata_list


def generate_files(line_list, args):
    files_list = ["<files>"]

    is_cpu_package = args.package_name in [
        "Microsoft.ML.OnnxRuntime",
        "Microsoft.ML.OnnxRuntime.OpenMP",
        "Microsoft.ML.OnnxRuntime.Training",
    ]
    is_mklml_package = args.package_name == "Microsoft.ML.OnnxRuntime.MKLML"
    is_cuda_gpu_package = args.package_name == "Microsoft.ML.OnnxRuntime.Gpu"
    is_cuda_gpu_win_sub_package = args.package_name == "Microsoft.ML.OnnxRuntime.Gpu.Windows"
    is_cuda_gpu_linux_sub_package = args.package_name == "Microsoft.ML.OnnxRuntime.Gpu.Linux"
    is_rocm_gpu_package = args.package_name == "Microsoft.ML.OnnxRuntime.ROCm"
    is_dml_package = args.package_name == "Microsoft.ML.OnnxRuntime.DirectML"
    is_windowsai_package = args.package_name == "Microsoft.AI.MachineLearning"
    is_snpe_package = args.package_name == "Microsoft.ML.OnnxRuntime.Snpe"
    is_qnn_package = args.package_name == "Microsoft.ML.OnnxRuntime.QNN"
    is_training_package = args.package_name in [
        "Microsoft.ML.OnnxRuntime.Training",
        "Microsoft.ML.OnnxRuntime.Training.Gpu",
    ]

    includes_winml = is_windowsai_package
    includes_directml = (is_dml_package or is_windowsai_package) and (
        args.target_architecture == "x64" or args.target_architecture == "x86"
    )

    is_windows_build = is_windows()

    nuget_dependencies = {}

    if is_windows_build:
        nuget_dependencies = {
            "mklml": "mklml.dll",
            "openmp": "libiomp5md.dll",
            "dnnl": "dnnl.dll",
            "providers_shared_lib": "onnxruntime_providers_shared.dll",
            "dnnl_ep_shared_lib": "onnxruntime_providers_dnnl.dll",
            "tensorrt_ep_shared_lib": "onnxruntime_providers_tensorrt.dll",
            "openvino_ep_shared_lib": "onnxruntime_providers_openvino.dll",
            "cuda_ep_shared_lib": "onnxruntime_providers_cuda.dll",
            "qnn_ep_shared_lib": "onnxruntime_providers_qnn.dll",
            "onnxruntime_perf_test": "onnxruntime_perf_test.exe",
            "onnx_test_runner": "onnx_test_runner.exe",
        }

        copy_command = "copy"
        runtimes_target = '" target="runtimes\\win-'
    else:
        nuget_dependencies = {
            "mklml": "libmklml_intel.so",
            "mklml_1": "libmklml_gnu.so",
            "openmp": "libiomp5.so",
            "dnnl": "libdnnl.so.1",
            "providers_shared_lib": "libonnxruntime_providers_shared.so",
            "dnnl_ep_shared_lib": "libonnxruntime_providers_dnnl.so",
            "tensorrt_ep_shared_lib": "libonnxruntime_providers_tensorrt.so",
            "openvino_ep_shared_lib": "libonnxruntime_providers_openvino.so",
            "cuda_ep_shared_lib": "libonnxruntime_providers_cuda.so",
            "rocm_ep_shared_lib": "libonnxruntime_providers_rocm.so",
            "onnxruntime_perf_test": "onnxruntime_perf_test",
            "onnx_test_runner": "onnx_test_runner",
        }

        copy_command = "cp"
        runtimes_target = '" target="runtimes\\linux-'

    if is_windowsai_package:
        runtimes_native_folder = "_native"
    else:
        runtimes_native_folder = "native"

    runtimes = f'{runtimes_target}{args.target_architecture}\\{runtimes_native_folder}"'

    # Process headers
    build_dir = "buildTransitive" if "Gpu" in args.package_name else "build"
    include_dir = f"{build_dir}\\native\\include"

    # Sub.Gpu packages do not include the onnxruntime headers
    if args.package_name != "Microsoft.ML.OnnxRuntime.Gpu":
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.sources_path, "include\\onnxruntime\\core\\session\\onnxruntime_*.h")
            + '" target="'
            + include_dir
            + '" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.sources_path, "include\\onnxruntime\\core\\framework\\provider_options.h")
            + '" target="'
            + include_dir
            + '" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.sources_path, "include\\onnxruntime\\core\\providers\\cpu\\cpu_provider_factory.h")
            + '" target="'
            + include_dir
            + '" />'
        )

    if is_training_package:
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(
                args.sources_path, "orttraining\\orttraining\\training_api\\include\\onnxruntime_training_*.h"
            )
            + '" target="build\\native\\include" />'
        )

    if args.execution_provider == "openvino":
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(
                args.sources_path, "include\\onnxruntime\\core\\providers\\openvino\\openvino_provider_factory.h"
            )
            + '" target="build\\native\\include" />'
        )

    if args.execution_provider == "tensorrt":
        files_list.append("<file src=" + '"' + '" target="build\\native\\include" />')

    if args.execution_provider == "dnnl":
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.sources_path, "include\\onnxruntime\\core\\providers\\dnnl\\dnnl_provider_options.h")
            + '" target="build\\native\\include" />'
        )

    if includes_directml:
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.sources_path, "include\\onnxruntime\\core\\providers\\dml\\dml_provider_factory.h")
            + '" target="build\\native\\include" />'
        )

    if includes_winml:
        # Add microsoft.ai.machinelearning headers
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.ort_build_path, args.build_config, "microsoft.ai.machinelearning.h")
            + '" target="build\\native\\include\\abi\\Microsoft.AI.MachineLearning.h" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.sources_path, "winml\\api\\dualapipartitionattribute.h")
            + '" target="build\\native\\include\\abi\\dualapipartitionattribute.h" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.ort_build_path, args.build_config, "microsoft.ai.machinelearning.native.h")
            + '" target="build\\native\\include\\Microsoft.AI.MachineLearning.Native.h" />'
        )
        # Add custom operator headers
        mlop_path = "onnxruntime\\core\\providers\\dml\\dmlexecutionprovider\\inc\\mloperatorauthor.h"
        files_list.append(
            "<file src=" + '"' + os.path.join(args.sources_path, mlop_path) + '" target="build\\native\\include" />'
        )
        # Process microsoft.ai.machinelearning.winmd
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.ort_build_path, args.build_config, "microsoft.ai.machinelearning.winmd")
            + '" target="winmds\\Microsoft.AI.MachineLearning.winmd" />'
        )
        # Process microsoft.ai.machinelearning.experimental.winmd
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.ort_build_path, args.build_config, "microsoft.ai.machinelearning.experimental.winmd")
            + '" target="winmds\\Microsoft.AI.MachineLearning.Experimental.winmd" />'
        )
        if args.target_architecture == "x64":
            interop_dll_path = "Microsoft.AI.MachineLearning.Interop\\net5.0-windows10.0.17763.0"
            interop_dll = interop_dll_path + "\\Microsoft.AI.MachineLearning.Interop.dll"
            files_list.append(
                "<file src="
                + '"'
                + os.path.join(args.native_build_path, interop_dll)
                + '" target="lib\\net5.0\\Microsoft.AI.MachineLearning.Interop.dll" />'
            )
            interop_pdb_path = "Microsoft.AI.MachineLearning.Interop\\net5.0-windows10.0.17763.0"
            interop_pdb = interop_pdb_path + "\\Microsoft.AI.MachineLearning.Interop.pdb"
            files_list.append(
                "<file src="
                + '"'
                + os.path.join(args.native_build_path, interop_pdb)
                + '" target="lib\\net5.0\\Microsoft.AI.MachineLearning.Interop.pdb" />'
            )

    if args.package_name == "Microsoft.ML.OnnxRuntime.Snpe" or args.package_name == "Microsoft.ML.OnnxRuntime.QNN":
        files_list.append(
            "<file src=" + '"' + os.path.join(args.native_build_path, "onnx_test_runner.exe") + runtimes + " />"
        )
        files_list.append(
            "<file src=" + '"' + os.path.join(args.native_build_path, "onnxruntime_perf_test.exe") + runtimes + " />"
        )

    if is_qnn_package:
        files_list.append("<file src=" + '"' + os.path.join(args.native_build_path, "QnnCpu.dll") + runtimes + " />")
        files_list.append("<file src=" + '"' + os.path.join(args.native_build_path, "QnnHtp.dll") + runtimes + " />")
        files_list.append("<file src=" + '"' + os.path.join(args.native_build_path, "QnnSaver.dll") + runtimes + " />")
        if args.target_architecture != "x64":
            files_list.append(
                "<file src=" + '"' + os.path.join(args.native_build_path, "QnnSystem.dll") + runtimes + " />"
            )
            files_list.append(
                "<file src=" + '"' + os.path.join(args.native_build_path, "QnnHtpPrepare.dll") + runtimes + " />"
            )
            files_list.append(
                "<file src=" + '"' + os.path.join(args.native_build_path, "QnnHtpV73Stub.dll") + runtimes + " />"
            )
            files_list.append(
                "<file src=" + '"' + os.path.join(args.native_build_path, "libQnnHtpV73Skel.so") + runtimes + " />"
            )
            files_list.append(
                "<file src=" + '"' + os.path.join(args.native_build_path, "libqnnhtpv73.cat") + runtimes + " />"
            )
            files_list.append(
                "<file src=" + '"' + os.path.join(args.native_build_path, "QnnHtpV68Stub.dll") + runtimes + " />"
            )
            files_list.append(
                "<file src=" + '"' + os.path.join(args.native_build_path, "libQnnHtpV68Skel.so") + runtimes + " />"
            )

    is_ado_packaging_build = False
    # Process runtimes
    # Process onnxruntime import lib, dll, and pdb
    # for Snpe android build
    if is_windows_build:
        nuget_artifacts_dir = Path(args.native_build_path) / "nuget-artifacts"
        # the winml package includes pdbs. for other packages exclude them.
        include_pdbs = includes_winml
        if nuget_artifacts_dir.exists():
            # Code path for ADO build pipeline, the files under 'nuget-artifacts' are
            # downloaded from other build jobs
            if is_cuda_gpu_package or is_cuda_gpu_win_sub_package or is_cuda_gpu_linux_sub_package:
                ep_list = ["tensorrt", "cuda", None]
            elif is_rocm_gpu_package:
                ep_list = ["rocm", None]
            else:
                ep_list = [None]
            for ep in ep_list:
                generate_file_list_for_ep(
                    nuget_artifacts_dir, ep, files_list, include_pdbs, is_training_package, args.package_name
                )
            is_ado_packaging_build = True
        else:
            # Code path for local dev build
            # for local dev build, gpu linux package is also generated for compatibility though it is not used
            if not is_cuda_gpu_linux_sub_package:
                files_list.append(
                    "<file src=" + '"' + os.path.join(args.native_build_path, "onnxruntime.lib") + runtimes + " />"
                )
                files_list.append(
                    "<file src=" + '"' + os.path.join(args.native_build_path, "onnxruntime.dll") + runtimes + " />"
                )
                if include_pdbs and os.path.exists(os.path.join(args.native_build_path, "onnxruntime.pdb")):
                    files_list.append(
                        "<file src=" + '"' + os.path.join(args.native_build_path, "onnxruntime.pdb") + runtimes + " />"
                    )
    else:
        ort_so = os.path.join(args.native_build_path, "libonnxruntime.so")
        if os.path.exists(ort_so):
            files_list.append(
                "<file src="
                + '"'
                + os.path.join(args.native_build_path, "libonnxruntime.so")
                + '" target="runtimes\\linux-'
                + args.target_architecture
                + '\\native" />'
            )

    if includes_winml:
        # Process microsoft.ai.machinelearning import lib, dll, and pdb
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, "microsoft.ai.machinelearning.lib")
            + runtimes_target
            + args.target_architecture
            + "\\_native"
            + '\\Microsoft.AI.MachineLearning.lib" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, "microsoft.ai.machinelearning.dll")
            + runtimes_target
            + args.target_architecture
            + "\\_native"
            + '\\Microsoft.AI.MachineLearning.dll" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, "microsoft.ai.machinelearning.pdb")
            + runtimes_target
            + args.target_architecture
            + "\\_native"
            + '\\Microsoft.AI.MachineLearning.pdb" />'
        )
    # Process execution providers which are built as shared libs
    if args.execution_provider == "tensorrt" and not is_ado_packaging_build:
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["providers_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["cuda_ep_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["tensorrt_ep_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )

    if args.execution_provider == "dnnl":
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["providers_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["dnnl_ep_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )

    if args.execution_provider == "rocm" or (is_rocm_gpu_package and not is_ado_packaging_build):
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["providers_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["rocm_ep_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )

    if args.execution_provider == "openvino":
        openvino_path = get_env_var("INTEL_OPENVINO_DIR")
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["providers_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["openvino_ep_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )

        if is_windows():
            dll_list_path = os.path.join(openvino_path, "runtime\\bin\\intel64\\Release\\")
            tbb_list_path = os.path.join(openvino_path, "runtime\\3rdparty\\tbb\\bin\\")
            for dll_element in os.listdir(dll_list_path):
                if dll_element.endswith("dll"):
                    files_list.append(
                        "<file src="
                        + '"'
                        + os.path.join(dll_list_path, dll_element)
                        + runtimes_target
                        + args.target_architecture
                        + '\\native" />'
                    )
            for tbb_element in os.listdir(tbb_list_path):
                if tbb_element.endswith("dll"):
                    files_list.append(
                        "<file src="
                        + '"'
                        + os.path.join(tbb_list_path, tbb_element)
                        + runtimes_target
                        + args.target_architecture
                        + '\\native" />'
                    )

    if args.execution_provider == "cuda" or (is_cuda_gpu_win_sub_package and not is_ado_packaging_build):
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["providers_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["cuda_ep_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )

    if args.execution_provider == "qnn" or (is_qnn_package and not is_ado_packaging_build):
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["providers_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, nuget_dependencies["qnn_ep_shared_lib"])
            + runtimes_target
            + args.target_architecture
            + '\\native" />'
        )

    # process all other library dependencies
    if is_cpu_package or is_cuda_gpu_package or is_dml_package or is_mklml_package:
        # Process dnnl dependency
        if os.path.exists(os.path.join(args.native_build_path, nuget_dependencies["dnnl"])):
            files_list.append(
                "<file src=" + '"' + os.path.join(args.native_build_path, nuget_dependencies["dnnl"]) + runtimes + " />"
            )

        # Process mklml dependency
        if os.path.exists(os.path.join(args.native_build_path, nuget_dependencies["mklml"])):
            files_list.append(
                "<file src="
                + '"'
                + os.path.join(args.native_build_path, nuget_dependencies["mklml"])
                + runtimes
                + " />"
            )

        if is_linux() and os.path.exists(os.path.join(args.native_build_path, nuget_dependencies["mklml_1"])):
            files_list.append(
                "<file src="
                + '"'
                + os.path.join(args.native_build_path, nuget_dependencies["mklml_1"])
                + runtimes
                + " />"
            )

        # Process libiomp5md dependency
        if os.path.exists(os.path.join(args.native_build_path, nuget_dependencies["openmp"])):
            files_list.append(
                "<file src="
                + '"'
                + os.path.join(args.native_build_path, nuget_dependencies["openmp"])
                + runtimes
                + " />"
            )

        # Some tools to be packaged in nightly debug build only, should not be released
        # These are copied to the runtimes folder for convenience of loading with the dlls
        # NOTE: nuget gives a spurious error on linux if these aren't in a separate directory to the library so
        #       we add them to a tools folder for that reason.
        if (
            args.is_release_build.lower() != "true"
            and args.target_architecture == "x64"
            and os.path.exists(os.path.join(args.native_build_path, nuget_dependencies["onnxruntime_perf_test"]))
        ):
            files_list.append(
                "<file src="
                + '"'
                + os.path.join(args.native_build_path, nuget_dependencies["onnxruntime_perf_test"])
                + runtimes[:-1]
                + "\\tools\\"
                + nuget_dependencies["onnxruntime_perf_test"]
                + '"'
                + " />"
            )

        if (
            args.is_release_build.lower() != "true"
            and args.target_architecture == "x64"
            and os.path.exists(os.path.join(args.native_build_path, nuget_dependencies["onnx_test_runner"]))
        ):
            files_list.append(
                "<file src="
                + '"'
                + os.path.join(args.native_build_path, nuget_dependencies["onnx_test_runner"])
                + runtimes[:-1]
                + "\\tools\\"
                + nuget_dependencies["onnx_test_runner"]
                + '"'
                + " />"
            )

    # Process props and targets files
    if is_windowsai_package:
        windowsai_src = "Microsoft.AI.MachineLearning"
        windowsai_props = "Microsoft.AI.MachineLearning.props"
        windowsai_targets = "Microsoft.AI.MachineLearning.targets"
        windowsai_native_props = os.path.join(args.sources_path, "csharp", "src", windowsai_src, windowsai_props)
        windowsai_rules = "Microsoft.AI.MachineLearning.Rules.Project.xml"
        windowsai_native_rules = os.path.join(args.sources_path, "csharp", "src", windowsai_src, windowsai_rules)
        windowsai_native_targets = os.path.join(args.sources_path, "csharp", "src", windowsai_src, windowsai_targets)
        build = f"{build_dir}\\native"
        files_list.append("<file src=" + '"' + windowsai_native_props + '" target="' + build + '" />')
        # Process native targets
        files_list.append("<file src=" + '"' + windowsai_native_targets + '" target="' + build + '" />')
        # Process rules
        files_list.append("<file src=" + '"' + windowsai_native_rules + '" target="' + build + '" />')
        # Process .net5.0 targets
        if args.target_architecture == "x64":
            interop_src = "Microsoft.AI.MachineLearning.Interop"
            interop_props = "Microsoft.AI.MachineLearning.props"
            interop_targets = "Microsoft.AI.MachineLearning.targets"
            windowsai_net50_props = os.path.join(args.sources_path, "csharp", "src", interop_src, interop_props)
            windowsai_net50_targets = os.path.join(args.sources_path, "csharp", "src", interop_src, interop_targets)
            files_list.append("<file src=" + '"' + windowsai_net50_props + '" target="build\\net5.0" />')
            files_list.append("<file src=" + '"' + windowsai_net50_targets + '" target="build\\net5.0" />')

    if (
        is_cpu_package
        or is_cuda_gpu_package
        or is_cuda_gpu_linux_sub_package
        or is_cuda_gpu_win_sub_package
        or is_rocm_gpu_package
        or is_dml_package
        or is_mklml_package
        or is_snpe_package
        or is_qnn_package
    ):
        # Process props file
        source_props = os.path.join(
            args.sources_path, "csharp", "src", "Microsoft.ML.OnnxRuntime", "targets", "netstandard", "props.xml"
        )
        target_props = os.path.join(
            args.sources_path,
            "csharp",
            "src",
            "Microsoft.ML.OnnxRuntime",
            "targets",
            "netstandard",
            args.package_name + ".props",
        )
        os.system(copy_command + " " + source_props + " " + target_props)
        files_list.append("<file src=" + '"' + target_props + '" target="' + build_dir + '\\native" />')
        if not is_snpe_package and not is_qnn_package:
            files_list.append("<file src=" + '"' + target_props + '" target="' + build_dir + '\\netstandard2.0"  />')
            files_list.append("<file src=" + '"' + target_props + '" target="' + build_dir + '\\netstandard2.1"  />')

        # Process targets file
        source_targets = os.path.join(
            args.sources_path, "csharp", "src", "Microsoft.ML.OnnxRuntime", "targets", "netstandard", "targets.xml"
        )
        target_targets = os.path.join(
            args.sources_path,
            "csharp",
            "src",
            "Microsoft.ML.OnnxRuntime",
            "targets",
            "netstandard",
            args.package_name + ".targets",
        )
        os.system(copy_command + " " + source_targets + " " + target_targets)
        files_list.append("<file src=" + '"' + target_targets + '" target="' + build_dir + '\\native"  />')
        if not is_snpe_package and not is_qnn_package:
            files_list.append("<file src=" + '"' + target_targets + '" target="' + build_dir + '\\netstandard2.0" />')
            files_list.append("<file src=" + '"' + target_targets + '" target="' + build_dir + '\\netstandard2.1"  />')

        # Process xamarin targets files
        if args.package_name == "Microsoft.ML.OnnxRuntime":
            net8_android_source_targets = os.path.join(
                args.sources_path,
                "csharp",
                "src",
                "Microsoft.ML.OnnxRuntime",
                "targets",
                "net8.0-android",
                "targets.xml",
            )
            net8_android_target_targets = os.path.join(
                args.sources_path,
                "csharp",
                "src",
                "Microsoft.ML.OnnxRuntime",
                "targets",
                "net8.0-android",
                args.package_name + ".targets",
            )

            net8_ios_source_targets = os.path.join(
                args.sources_path, "csharp", "src", "Microsoft.ML.OnnxRuntime", "targets", "net8.0-ios", "targets.xml"
            )
            net8_ios_target_targets = os.path.join(
                args.sources_path,
                "csharp",
                "src",
                "Microsoft.ML.OnnxRuntime",
                "targets",
                "net8.0-ios",
                args.package_name + ".targets",
            )

            net8_maccatalyst_source_targets = os.path.join(
                args.sources_path,
                "csharp",
                "src",
                "Microsoft.ML.OnnxRuntime",
                "targets",
                "net8.0-maccatalyst",
                "_._",
            )

            net8_maccatalyst_target_targets = os.path.join(
                args.sources_path, "csharp", "src", "Microsoft.ML.OnnxRuntime", "targets", "net8.0-maccatalyst", "_._"
            )

            os.system(copy_command + " " + net8_android_source_targets + " " + net8_android_target_targets)
            os.system(copy_command + " " + net8_ios_source_targets + " " + net8_ios_target_targets)
            os.system(copy_command + " " + net8_maccatalyst_source_targets + " " + net8_maccatalyst_target_targets)

            files_list.append(
                "<file src=" + '"' + net8_android_target_targets + '" target="build\\net8.0-android31.0" />'
            )
            files_list.append(
                "<file src=" + '"' + net8_android_target_targets + '" target="buildTransitive\\net8.0-android31.0" />'
            )

            files_list.append("<file src=" + '"' + net8_ios_target_targets + '" target="build\\net8.0-ios15.4" />')
            files_list.append(
                "<file src=" + '"' + net8_ios_target_targets + '" target="buildTransitive\\net8.0-ios15.4" />'
            )
            files_list.append(
                "<file src=" + '"' + net8_maccatalyst_target_targets + '" target="build\\net8.0-maccatalyst14.0" />'
            )
            files_list.append(
                "<file src="
                + '"'
                + net8_maccatalyst_target_targets
                + '" target="buildTransitive\\net8.0-maccatalyst14.0" />'
            )

        # Process Training specific targets and props
        if args.package_name == "Microsoft.ML.OnnxRuntime.Training":
            net8_android_source_targets = os.path.join(
                args.sources_path,
                "csharp",
                "src",
                "Microsoft.ML.OnnxRuntime",
                "targets",
                "net8.0-android",
                "targets.xml",
            )
            net8_android_target_targets = os.path.join(
                args.sources_path,
                "csharp",
                "src",
                "Microsoft.ML.OnnxRuntime",
                "targets",
                "net8.0-android",
                args.package_name + ".targets",
            )

            os.system(copy_command + " " + net8_android_source_targets + " " + net8_android_target_targets)
            files_list.append(
                "<file src=" + '"' + net8_android_target_targets + '" target="build\\net8.0-android31.0" />'
            )
            files_list.append(
                "<file src=" + '"' + net8_android_target_targets + '" target="buildTransitive\\net8.0-android31.0" />'
            )

    # README
    files_list.append(
        "<file src=" + '"' + os.path.join(args.sources_path, "tools/nuget/nupkg.README.md") + '" target="README.md" />'
    )

    # Process License, ThirdPartyNotices, Privacy
    files_list.append("<file src=" + '"' + os.path.join(args.sources_path, "LICENSE") + '" target="LICENSE" />')
    files_list.append(
        "<file src="
        + '"'
        + os.path.join(args.sources_path, "ThirdPartyNotices.txt")
        + '" target="ThirdPartyNotices.txt" />'
    )
    files_list.append(
        "<file src=" + '"' + os.path.join(args.sources_path, "docs", "Privacy.md") + '" target="Privacy.md" />'
    )
    files_list.append(
        "<file src="
        + '"'
        + os.path.join(args.sources_path, "ORT_icon_for_light_bg.png")
        + '" target="ORT_icon_for_light_bg.png" />'
    )
    if is_qnn_package:
        files_list.append(
            "<file src="
            + '"'
            + os.path.join(args.native_build_path, "Qualcomm AI Hub Proprietary License.pdf")
            + '" target="Qualcomm AI Hub Proprietary License.pdf" />'
        )
    files_list.append("</files>")

    line_list += files_list


def generate_nuspec(args):
    lines = ['<?xml version="1.0"?>']
    lines.append("<package>")
    generate_metadata(lines, args)
    generate_files(lines, args)
    lines.append("</package>")
    return lines


def is_windows():
    return sys.platform.startswith("win")


def is_linux():
    return sys.platform.startswith("linux")


def is_macos():
    return sys.platform.startswith("darwin")


def validate_platform():
    if not (is_windows() or is_linux() or is_macos()):
        raise Exception("Native Nuget generation is currently supported only on Windows, Linux, and MacOS")


def validate_execution_provider(execution_provider):
    if is_linux():
        if not (
            execution_provider == "None"
            or execution_provider == "dnnl"
            or execution_provider == "cuda"
            or execution_provider == "tensorrt"
            or execution_provider == "openvino"
            or execution_provider == "rocm"
        ):
            raise Exception(
                "On Linux platform nuget generation is supported only "
                "for cpu|cuda|dnnl|tensorrt|openvino|rocm execution providers."
            )


def main():
    # Parse arguments
    args = parse_arguments()

    validate_platform()

    validate_execution_provider(args.execution_provider)

    if args.is_release_build.lower() != "true" and args.is_release_build.lower() != "false":
        raise Exception("Only valid options for IsReleaseBuild are: true and false")

    # Generate nuspec
    lines = generate_nuspec(args)

    # Create the nuspec needed to generate the Nuget
    print(f"nuspec_name: {args.nuspec_name}")
    with open(os.path.join(args.native_build_path, args.nuspec_name), "w") as f:
        for line in lines:
            # Uncomment the printing of the line if you need to debug what's produced on a CI machine
            # print(line)
            f.write(line)
            f.write("\n")


if __name__ == "__main__":
    sys.exit(main())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import argparse
import importlib.util
import os
import shutil

import triton


def compile(function_table, out_dir):
    def compile_one(func, sig, **kwargs):
        ret = triton.compile(func, signature=sig, **kwargs)
        return ret

    metadata = []
    for func_desc in function_table:
        name = func_desc["name"]
        group = func_desc["group"]
        sig = func_desc["sig"]
        func = func_desc["func"]
        kwargs = func_desc["kwargs"]

        # print("compile func: ", func_desc)

        ret = compile_one(func, sig, **kwargs)

        compile_res = {}
        compile_res["name"] = name
        compile_res["group"] = group
        compile_res["func_name"] = ret.metadata["name"]
        compile_res["num_warps"] = ret.metadata["num_warps"]
        compile_res["shared"] = ret.metadata["shared"]
        if "constants" in kwargs:
            compile_res["constants"] = kwargs["constants"]

        # move tmp kernel file into current dir
        if "hsaco_path" in ret.asm and os.path.exists(ret.asm["hsaco_path"]):
            # is rocm
            lib_name = f"{name}.hsaco"
            shutil.copyfile(ret.asm["hsaco_path"], f"{out_dir}/{lib_name}")
        elif "cubin" in ret.asm:
            # is cuda
            lib_name = f"{name}.cubin"
            # need to write cubin into file
            with open(f"{out_dir}/{lib_name}", "wb") as fp:
                fp.write(ret.asm["cubin"])
        else:
            raise Exception("not find rocm or cuda compiled kernel")

        compile_res["lib_file"] = lib_name
        metadata.append(compile_res)

    return metadata


def convert_lib_to_obj(lib_file, out_dir):
    obj_file = lib_file.split(".")[0] + ".o"
    command = f"cd {out_dir}; objcopy -I binary -O elf64-x86-64 -B i386:x86-64 {lib_file} {obj_file}; cd -"
    ret = os.system(command)

    if ret != 0:
        raise Exception(f"exec convert command: {command} failed.")
    # check file exist
    if not os.path.exists(f"{out_dir}/{obj_file}"):
        raise Exception(f"the output file not exist, after exec comamnd: {command}")

    return obj_file


def archive_obj_files(obj_files, out_dir, out_obj_file):
    obj_files = " ".join(obj_files)
    command = f"cd {out_dir}; ar rcs {out_obj_file} {obj_files}; cd -"
    ret = os.system(command)

    if ret != 0:
        raise Exception(f"exec convert command: {command} failed.")
    # check file exist
    if not os.path.exists(f"{out_dir}/{out_obj_file}"):
        raise Exception(f"the output file not exist, after exec comamnd: {command}")


def convert_and_save(metadata, header_file, out_dir, out_obj_file):
    c_metadata = []
    binary_files = []
    for m in metadata:
        meta_ele = []
        obj_file = convert_lib_to_obj(m["lib_file"], out_dir)
        binary_files.append(obj_file)

        lib_name = m["lib_file"].replace(".", "_")
        meta_ele.append(f'"_binary_{lib_name}_start"')
        meta_ele.append(f'"{m["func_name"]}"')
        meta_ele.append(f'"{m["group"]}"')
        meta_ele.append(f'"{m["name"]}"')
        meta_ele.append(str(m["num_warps"]))
        meta_ele.append(str(m["shared"]))

        # convert constants
        constants = []
        for k, v in m["constants"].items():
            constants.append(f'{{ "{k}", {v!s}}}')
        meta_ele.append(f"{{ {', '.join(constants)} }}")

        c_metadata.append(f"{{ {', '.join(meta_ele)} }}")

    archive_obj_files(binary_files, out_dir, out_obj_file)

    code = f"""
#include <unordered_map>

struct _TritonKernelInfo {{
  const char* name_start;
  const char* func_name;
  const char* group_name;
  const char* name;
  int num_warps;
  int shared;
  std::unordered_map<std::string, int> constants;
}};

const _TritonKernelInfo kernel_infos[] = {{
  {", ".join(c_metadata)},
}};
    """

    with open(header_file, "w") as fp:
        fp.write(code)


def main(args):
    out_obj_file = args.obj_file
    out_dir = os.path.dirname(out_obj_file)
    out_obj_file = os.path.basename(out_obj_file)
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    metadata = []
    print("[triton kernel] start compile triton kernel.")
    for i, f in enumerate(args.script_files):
        # import module in f, and call function
        spec = importlib.util.spec_from_file_location(f"module_{i}", f)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        func_tb = module.get_function_table()
        m = compile(func_tb, out_dir)
        metadata.extend(m)

    print("[triton kernel] compile triton kernel done.")

    # save metadata into header file
    convert_and_save(metadata, args.header, out_dir, out_obj_file)
    print("[triton kernel] save into file done.")


def get_arges():
    parser = argparse.ArgumentParser(description="PyTorch Template Finetune Example")
    parser.add_argument(
        "--header", type=str, default="triton_kernel_infos.h", help="the header file that should be generated."
    )
    parser.add_argument("--ort_root", type=str, default="onnxruntime", help="the root dir of onnxruntime.")
    parser.add_argument("--script_files", type=str, nargs="+", help="the root dir of onnxruntime.")
    parser.add_argument("--obj_file", type=str, default="triton_kernel_infos.a", help="output target object files.")

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = get_arges()
    main(args)
